/****************************************************************************
*   PROJECT: Squeak port for Win32 (NT / Win95)
*   FILE:    sqWin32Window.c
*   CONTENT: Window management
*
*   AUTHOR:  Andreas Raab (ar)
*   ADDRESS: University of Magdeburg, Germany
*   EMAIL:   raab@isg.cs.uni-magdeburg.de
*   RCSID:   $Id: sqWin32Window.c,v 1.12 2001/02/28 08:58:44 anonymous Exp $
*
*   NOTES:
*    1) Currently supported Squeak color depths include 1,4,8,16,32 bits
*    2) To speed up drawing a slight update delay has been added (Toggle with F2)
*    3) The modifier keys are mapped as follows:
*
*        Mac    |  Win32
*       --------------------
*       Shift   -> Shift
*       Ctrl    -> Ctrl
*       Command -> Left ALT
*       Option  -> Right ALT
*
*****************************************************************************/
#include <windows.h>
#include <commdlg.h>
#include <excpt.h>

#include "sq.h"
#include "sqWin32Prefs.h"

#if _WIN32_WCE >= 300
#  include <aygshell.h>
#  include <sipapi.h>
#  ifdef USE_GAPI
#    include "gx.h"
#  endif
#endif

#ifndef NO_RCSID
static TCHAR RCSID[]= TEXT("$Id: sqWin32Window.c,v 1.12 2001/02/28 08:58:44 anonymous Exp $");
#endif

/****************************************************************************/
/* General Squeak declarations and definitions                              */
/****************************************************************************/

/*** Variables -- Imported from Virtual Machine ***/
extern "C" {
extern int interruptPending;
extern int interruptCheckCounter;
extern int interruptKeycode;
extern int fullScreenFlag;
extern int deferDisplayUpdates; /* Is the Interpreter doing defered updates for us?! */
}

/*** Variables -- image and path names ***/
#define IMAGE_NAME_SIZE MAX_PATH

char imageName[MAX_PATH+1];		  /* full path and name to image */
TCHAR imagePath[MAX_PATH+1];	  /* full path to image */
TCHAR vmPath[MAX_PATH+1];		    /* full path to interpreter's directory */
TCHAR vmName[MAX_PATH+1];		    /* name of the interpreter's executable */

int		 savedWindowSize= 0;	/* initial size of window */

/*** Variables -- Event Recording ***/
int inputSemaphoreIndex = 0;/* if non-zero the event semaphore index */
extern "C" {
POINT mousePosition;		/* position at last PointerMotion event */
}
int   mouseWord;			/* Input word for Squeak */
int   buttonState = 0;		/* mouse button and modifier state when mouse
							   button went down or 0 if not pressed */

#define KEYBUF_SIZE 64
int keyBuf[KEYBUF_SIZE];	/* circular buffer */
int keyBufGet = 0;			/* index of next item of keyBuf to read */
int keyBufPut = 0;			/* index of next item of keyBuf to write */
int keyBufOverflows = 0;	/* number of characters dropped */

/*** Win32-related Variables (declared in sqWin32.h) ***/
HWND stWindow = NULL;      /*	the squeak window */
HWND browserWindow = NULL; /* The browser window */
HINSTANCE hInstance;	     /*	the instance of squeak running */
HCURSOR currentCursor=0;	 /*	current cursor displayed by squeak */
HPALETTE palette;	         /*	the palette (might be unused) */
LOGPALETTE *logPal;	       /*	the logical palette definition */
BITMAPINFO *bmi1;	         /*	1 bit depth bitmap info */
BITMAPINFO *bmi4;	         /*	4 bit depth bitmap info */
BITMAPINFO *bmi8;	         /*	8 bit depth bitmap info */
BITMAPINFO *bmi16;	       /*	16 bit depth bitmap info */
BITMAPINFO *bmi32;	       /*	32 bit depth bitmap info */
BOOL fWindows95;           /* Are we running on Win95 or NT? */
BOOL fHasFocus = 0;        /* True if Squeak has the input focus */

/* Preference values */
BOOL fDeferredUpdate = 1; /* I prefer the deferred update*/
BOOL fShowConsole = 0;    /* do we show the console window?*/
BOOL fDynamicConsole = 1; /* Should we show the console if any errors occur? */
BOOL fShowAllocations = 0; /* Show allocation activity */
BOOL fReduceCPUUsage = 1; /* Should we reduce CPU usage? */
BOOL fReduceCPUInBackground = 0; /* Should we reduce CPU usage when not active? */
BOOL fUseDirectSound = 1; /* Do we use DirectSound?! */

BOOL f1ButtonMouse = 0;   /* Should we use a 1 button mouse mapping? */
BOOL f3ButtonMouse = 0;   /* Should we use a real 3 button mouse mapping? */

/* Startup options */
BOOL  fHeadlessImage = 0;      /* Do we run headless? */
BOOL  fRunService = 0;         /* Do we run as NT service? */
DWORD dwMemorySize = 0;        /* How much memory do we use? */
BOOL  fBrowserMode = 0;        /* Are we running in a web browser? */

HANDLE vmSemaphoreMutex = 0; /* the mutex for synchronization */
HANDLE vmWakeUpEvent = 0;      /* wake up interpret() from sleep */

/* variables for cached display */
RECT updateRect;		     /*	the rectangle to update */
HRGN updateRgn;	     	     /*	the region to update (more accurate) */
BOOL updateRightNow;	     /*	update flag */
HWND  consoleWindow;       /* console */

/* variables for DirectX support */
RECT stWindowRect;			/* Client rectangle in screen coordinates */

#ifndef NO_PRINTER
/* printer settings */
PRINTDLG printValues;
#endif

#ifndef NO_WHEEL_MOUSE
UINT g_WM_MOUSEWHEEL = 0;	/* RvL: 1999-04-19 The message we receive from wheel mices */
#endif

/* misc declarations */
int recordMouseEvent(MSG *msg);
int recordKeyboardEvent(MSG *msg);

extern "C" {
extern int byteSwapped(int);
extern int convertToSqueakTime(SYSTEMTIME);
extern void sqNetPollSync();
void deleteBackingStoreFile();
}
int recordMouseDown(WPARAM, LPARAM);
int recordModifierButtons();
int recordKeystroke(UINT,WPARAM,LPARAM);
int recordVirtualKey(UINT,WPARAM,LPARAM);
void recordMouse(void);
void SetSystemTrayIcon(BOOL on);

static sqInputEvent *nextEventPut(void);

/* UNICODE stuff */
const TCHAR U_SLASH[] = TEXT("/");
const TCHAR U_BACKSLASH[] = TEXT("\\");

#ifdef USE_GAPI
int ioCESetFullScreen(int fullScreen);
void OpenDisplayOnIPAQ();
void CloseDisplayOnIPAQ();
void StartWritingDisplayOnIPAQ();
void StopWritingDisplayOnIPAQ();

int drawdisp_max_line;
int headMargin;
int borderWidthX;
int borderWidthY;

struct GXDisplayProperties g_gxdp;
struct GXKeyList g_gxkl;

#define bytesPerLine(width, depth)	((((width)*(depth) + 31) >> 5) << 2)
#define bytesPerLineRD(width, depth)	((((width)*(depth)) >> 5) << 2)

unsigned short stColors16[256];
#endif


/****************************************************************************/
/*                      Synchronization functions                           */
/****************************************************************************/

/* NOTE: Why do we need this? When running multi-threaded code such as in
         the networking code and in midi primitives
         we will signal the interpreter several semaphores. If one of those
         threads gets interrupted by any other thread trying to signal
         the interpreter we might have a problem. Even with the current
         synchronization we live somewhat dangerous, since the main thread
         does not use this synchronization when accessing the external
         semaphores. */

int synchronizedSignalSemaphoreWithIndex(int semaIndex)
{ int result;

  /* wait until we have access */
  WaitForSingleObject(vmSemaphoreMutex, INFINITE);
  /* do our job */
  result = signalSemaphoreWithIndex(semaIndex);
  /* wake up interpret() if sleeping */
  SetEvent(vmWakeUpEvent);
  /* and release access */
  ReleaseMutex(vmSemaphoreMutex);
  return result;
}



/****************************************************************************/
/*                   Message Processing                                     */
/****************************************************************************/

/* The last dispatched event. It is used for the event processing mechanism. */
static MSG *lastMessage = NULL;

extern "C" {
extern messageHook firstMessageHook;
extern messageHook preMessageHook;
}

/* main window procedure */
LRESULT CALLBACK MainWndProc (HWND hwnd,
                              UINT message,
                              WPARAM wParam,
                              LPARAM lParam)
{ PAINTSTRUCT ps;

  /* Intercept any messages if wanted */
  if(preMessageHook)
    if((*preMessageHook)(hwnd, message,wParam, lParam))
       return 1;

#ifndef NO_WHEEL_MOUSE
  /* RvL 1999-04-19 00:23 
     MOUSE WHEELING START */
  if( WM_MOUSEWHEEL == message || g_WM_MOUSEWHEEL == message ) {
    /* Record mouse wheel msgs as CTRL-Up/Down */
    short zDelta = (short) HIWORD(wParam);
    if(inputSemaphoreIndex) {
      sqKeyboardEvent *evt = (sqKeyboardEvent*) nextEventPut();
      evt->type = EventTypeKeyboard;
      evt->timeStamp = lastMessage->time;
      evt->charCode = (zDelta > 0) ? 30 : 31;
      evt->pressCode = EventKeyChar;
      evt->modifiers = CtrlKeyBit;
      evt->reserved1 = evt->reserved2 = 0;
    } else {
      buttonState = 64;
      if (zDelta < 0) {
	recordVirtualKey(message,VK_DOWN,lParam);
      } else {
	recordVirtualKey(message,VK_UP,lParam);
      }
    }
    return 1;	
  }
  /* MOUSE WHEELING END */
#endif

  switch(message) {
  case WM_CREATE:
    break;
  case WM_SYSCOMMAND:
  case WM_COMMAND: {
    int cmd = wParam & 0xFFF0;
    if(cmd >= ID_PREF_FIRST && cmd <= ID_PREF_LAST) {
      HandlePrefsMenu(cmd);
      break;
    }
#if !defined(_WIN32_WCE)
    if(cmd == SC_MINIMIZE) {
      if(fHeadlessImage) ShowWindow(stWindow, SW_HIDE);
      else return DefWindowProc(hwnd, message, wParam, lParam);
      break;
    }
#endif /* defined(_WIN32_WCE) */
    if(cmd == SC_CLOSE) {
      /* do not attempt to close the window directly ! */
      if(MessageBox(stWindow,TEXT("Quit Squeak without saving?"),
		    TEXT("Squeak"),MB_YESNO) != IDYES) return 0;
      DestroyWindow(stWindow);
      exit(1);
      break;
    }
    return DefWindowProc(hwnd,message,wParam,lParam);
    break;
  }
  /*  mousing */
  case WM_MOUSEMOVE: 
    if(inputSemaphoreIndex) {
      recordMouseEvent(lastMessage);
      break;
    }
    /* state based stuff */
    mousePosition.x = LOWORD(lParam);
    mousePosition.y = HIWORD(lParam);
    break;
  case WM_LBUTTONDOWN:
  case WM_RBUTTONDOWN:
  case WM_MBUTTONDOWN:
    if(GetFocus() != stWindow) SetFocus(stWindow);
/* #ifndef _WIN32_WCE */
    SetCapture(stWindow); /* capture mouse input */
/* #endif */
    if(inputSemaphoreIndex) {
      recordMouseEvent(lastMessage);
      break;
    }
    /* state based stuff */
    mousePosition.x = LOWORD(lParam);
    mousePosition.y = HIWORD(lParam);
    /* check for console focus */
    recordMouseDown(wParam, lParam);
    recordModifierButtons();
#ifndef _WIN32_WCE
    /* capture the mouse as long as the button is pressed so we can scroll outside */
    SetCapture(stWindow);
#endif
    break;
  case WM_LBUTTONUP:
  case WM_RBUTTONUP:
  case WM_MBUTTONUP:
    if(GetFocus() != stWindow) SetFocus(stWindow);
    ReleaseCapture(); /* release mouse capture */
    if(inputSemaphoreIndex) {
      recordMouseEvent(lastMessage);
      break;
    }
    /* state based stuff */
    mousePosition.x = LOWORD(lParam);
    mousePosition.y = HIWORD(lParam);
    /* check for console focus */
    if(GetFocus() != stWindow) SetFocus(stWindow);
    recordMouseDown(wParam,lParam);
    recordModifierButtons();
#ifndef _WIN32_WCE
    /* release capture */
    ReleaseCapture();
#endif
    break;
    /* virtual key codes */
  case WM_KEYDOWN:
  case WM_SYSKEYDOWN:
    if(GetFocus() == consoleWindow) 
      return DefWindowProc(hwnd, message, wParam, lParam);
    if(inputSemaphoreIndex) {
      recordKeyboardEvent(lastMessage);
      if(wParam == VK_F2) {
	TrackPrefsMenu();
      }
      if(wParam == VK_F4) {
	return DefWindowProc(hwnd, message, wParam, lParam);
      }
      break;
    }
    /* state based stuff */
    recordModifierButtons();
    if(!recordVirtualKey(message,wParam,lParam))
      return DefWindowProc(hwnd,message,wParam,lParam);
    break;
  case WM_KEYUP:
  case WM_SYSKEYUP:
    if(GetFocus() == consoleWindow) 
      return DefWindowProc(hwnd, message, wParam, lParam);
    if(inputSemaphoreIndex) {
      recordKeyboardEvent(lastMessage);
      break;
    }
    /* state based stuff */
    recordModifierButtons();
    break;
    /* character codes */
  case WM_CHAR:
  case WM_SYSCHAR:
    if(GetFocus() == consoleWindow) 
      return DefWindowProc(hwnd, message, wParam, lParam);
    if(inputSemaphoreIndex) {
      recordKeyboardEvent(lastMessage);
      break;
    }
    /* state based stuff */
    recordModifierButtons();
    recordKeystroke(message,wParam,lParam);
    break;
  case WM_DEADCHAR:
  case WM_SYSDEADCHAR:
    break;
    /* expose events */
  case WM_PAINT:
    /* retrieve update region _before_ calling BeginPaint() !!! */
#ifdef USE_GAPI
    if (GetForegroundWindow() == hwnd) {
       // should probably be in ioShowDisplay()
#endif
      if (GetUpdateRect(hwnd, NULL, FALSE)) {
        /* An application should call the GetUpdateRect function to determine
         * whether the window has an update region. If GetUpdateRect returns
         * zero, the application should not call the BeginPaint and EndPaint functions. */

        GetUpdateRgn(hwnd,updateRgn,FALSE);
        BeginPaint(hwnd,&ps);
        /* store the rectangle required for image bit reversal */
        updateRect = ps.rcPaint;
        EndPaint(hwnd,&ps);
        /* Avoid repaint unless absolutely necessary (null update rects are quite common) */
        // if (!IsRectEmpty(&updateRect)) {
           /* force redraw the next time ioShowDisplay() is called */
           updateRightNow = TRUE;
           fullDisplayUpdate();  /* this makes VM call ioShowDisplay */
        // }
      }
#ifdef USE_GAPI
    }
#endif
    break;
  case WM_SIZE:
    if(hwnd == stWindow) {
      /* Adjust the console window */
      GetClientRect(stWindow,&stWindowRect);
#if !defined(_WIN32_WCE)
      MoveWindow(consoleWindow, 0, 
		 stWindowRect.bottom-100, stWindowRect.right, 100, 1);
#endif
      /* Record the global stWindowRect for DirectX */
      MapWindowPoints(stWindow, NULL, (LPPOINT)&stWindowRect, 2);
    }
    else return DefWindowProc(hwnd,message,wParam,lParam);
    break;
  case WM_MOVE:
    if(hwnd == stWindow) {
      /* Record the global stWindowRect for DirectX */
      GetClientRect(stWindow,&stWindowRect);
      MapWindowPoints(stWindow, NULL, (LPPOINT) &stWindowRect, 2);
    }
    else return DefWindowProc(hwnd,message,wParam,lParam);
    break;
#if defined(_WIN32_WCE)
  case WM_CLOSE:
    if(MessageBox(stWindow,TEXT("Quit Squeak without saving?"),TEXT("Squeak"),MB_YESNO) == IDYES)
      {
	DestroyWindow(stWindow);
	ExitThread(1);
	return 0;
      }
    else return 1;
    break;
#  if 0
     /* All erasing is done in the display function 
	and the window is always full screen*/
  case WM_ERASEBKGND:
    return TRUE;
#  endif
#endif

#if !defined(_WIN32_WCE)
    /* Don't change the cursor or system tray on WinCE */
    /* cursor redraw */
  case WM_SETCURSOR:
    /* keep currentCursor */
    if((LOWORD(lParam) == HTCLIENT) && currentCursor) {
      SetCursor(currentCursor);
      break;
    }
    else return DefWindowProc(hwnd,message,wParam,lParam);
  case WM_USER+42:
    /* system tray notification */
    if(wParam != (UINT)hInstance) return 0;
    /* if right button, show system menu */
    /* if(lParam == WM_RBUTTONUP)
       TrackPrefsMenu(GetSystemMenu(stWindow,0)); */
    /* if double clicked, show main window */
    if(lParam == WM_LBUTTONDBLCLK) {
      if(!IsWindowVisible(stWindow))
	ShowWindow(stWindow, SW_SHOW);
      BringWindowToTop(stWindow);
    }
    return 0;
#endif /* !defined(_WIN32_WCE) */
    /* Focus handling */
  case WM_SETFOCUS:
    fHasFocus = 1;
# ifdef USE_GAPI
    GXResume();
# endif
    return DefWindowProc(hwnd,message,wParam,lParam);
  case WM_KILLFOCUS:
    fHasFocus = 0;
# ifdef USE_GAPI
    GXSuspend();
# endif
    return DefWindowProc(hwnd,message,wParam,lParam);
  case WM_WINDOWPOSCHANGED:
    if(hwnd == stWindow)
      {
#if _WIN32_WCE >= 300
	{
	  LPWINDOWPOS lpwp = (LPWINDOWPOS) lParam;
	  if (lpwp->hwndInsertAfter == HWND_TOP) {
	    SHFullScreen(stWindow, SHFS_HIDETASKBAR);
	  }
	}
#elif _WIN32_WCE >= 210
	{
	  LPWINDOWPOS lpwp = (LPWINDOWPOS) lParam;
	  RECT rect;
	  /* printf("WPC x %d y %d cx %d cy %d flags 0x%x\n", lpwp->x, lpwp->y, lpwp->cx, lpwp->cy, lpwp->flags); */
	  rect.left = lpwp->x;
	  rect.right = lpwp->x + lpwp->cx;
	  rect.top = lpwp->y;
	  rect.bottom = lpwp->y + lpwp->cy;
	  InvalidateRect(hwnd, &rect, TRUE);
	  return 0;     /* avoid DefWindowProc */
	}
#endif
      }
    return DefWindowProc(hwnd,message,wParam,lParam);

  default:
    /* Unprocessed messages may be processed outside the current
       module. If firstMessageHook is non-NULL and returns a non
       zero value, the message has been successfully processed */
    if(firstMessageHook)
      if((*firstMessageHook)(hwnd, message, wParam, lParam))
	return 1;
    return DefWindowProc(hwnd,message,wParam,lParam);
  }
  return 1;
}

/****************************************************************************/
/*                      Timer Setup                                         */
/****************************************************************************/
/****************************************************************************/
/* Windows CE does not support waiting for semaphores, change notification  */
/* objects, console input, and timers, as does the Win32 version.           */
/* Also, it does not support waiting for process, thread, and mutex events. */
/****************************************************************************/

static DWORD dwTimerPeriod;
#include <mmsystem.h>
void SetupTimer()
{
#if defined(_WIN32_WCE)
  dwTimerPeriod = 0;
#else /* defined(_WIN32_WCE) */
  TIMECAPS tCaps;

  dwTimerPeriod = 0;
  if(timeGetDevCaps(&tCaps,sizeof(tCaps)) != 0)
    return;
  dwTimerPeriod = tCaps.wPeriodMin;
  if(timeBeginPeriod(dwTimerPeriod) != 0)
    return;
#endif /* defined(_WIN32_WCE) */
}

void ReleaseTimer()
{
#if !defined(_WIN32_WCE)
  timeEndPeriod(dwTimerPeriod);
#endif /* !defined(_WIN32_WCE) */
}

/****************************************************************************/
/*                     Printer Setup                                        */
/****************************************************************************/

#ifndef NO_PRINTER
void SetDefaultPrinter()
{
  printValues.Flags = PD_PRINTSETUP;
  PrintDlg(&printValues);
}

void SetupPrinter()
{
  ZeroMemory(&printValues, sizeof(printValues));
  printValues.lStructSize = sizeof(PRINTDLG);
  printValues.hInstance = hInstance;
  printValues.nMinPage = 1;
  printValues.nMaxPage = 1;
  printValues.nFromPage = 1;
  printValues.nToPage = 1;
  printValues.hwndOwner = stWindow;

  /* fetch default DEVMODE/DEVNAMES */
  printValues.Flags = PD_RETURNDEFAULT;
  PrintDlg(&printValues);
}
#endif


/****************************************************************************/
/*                   Color and Bitmap setup                                 */
/****************************************************************************/
#ifndef NO_STANDARD_COLORS
static void SetColorEntry(int index, int red, int green, int blue)
{
#ifdef USE_GAPI
  unsigned short r, g, b;
  unsigned short w;

  r = red;
  g = green;
  b = blue;

  w = (((r>>11)<<11) |
       ((g>>10)<<5) | 
       ((b>>11)));

  if (index < 256) {
    stColors16[index] = w;
  }
#endif

  logPal->palPalEntry[index].peRed = red / 256;
  logPal->palPalEntry[index].peGreen = green / 256;
  logPal->palPalEntry[index].peBlue = blue / 256;
  logPal->palPalEntry[index].peFlags = 0;
}

/* Generic color maps and bitmap info headers 1,4,8,16,32 bits per pixel */
void SetupPixmaps(void)
{ int i;

  logPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) + 255 * sizeof(PALETTEENTRY));
  logPal->palVersion = 0x300;
  logPal->palNumEntries = 256;

  /* 1-bit colors (monochrome) */
  SetColorEntry(0, 65535, 65535, 65535);	/* white or transparent */
  SetColorEntry(1,     0,     0,     0);	/* black */

  /* additional colors for 2-bit color */
  SetColorEntry(2, 65535, 65535, 65535);	/* opaque white */
  SetColorEntry(3, 32768, 32768, 32768);	/* 1/2 gray */

  /* additional colors for 4-bit color */
  SetColorEntry( 4, 65535,     0,     0);	/* red */
  SetColorEntry( 5,     0, 65535,     0);	/* green */
  SetColorEntry( 6,     0,     0, 65535);	/* blue */
  SetColorEntry( 7,     0, 65535, 65535);	/* cyan */
  SetColorEntry( 8, 65535, 65535,     0);	/* yellow */
  SetColorEntry( 9, 65535,     0, 65535);	/* magenta */
  SetColorEntry(10,  8192,  8192,  8192);	/* 1/8 gray */
  SetColorEntry(11, 16384, 16384, 16384);	/* 2/8 gray */
  SetColorEntry(12, 24576, 24576, 24576);	/* 3/8 gray */
  SetColorEntry(13, 40959, 40959, 40959);	/* 5/8 gray */
  SetColorEntry(14, 49151, 49151, 49151);	/* 6/8 gray */
  SetColorEntry(15, 57343, 57343, 57343);	/* 7/8 gray */


  /* additional colors for 8-bit color */
  /* 24 more shades of gray (does not repeat 1/8th increments) */
  SetColorEntry(16,  2048,  2048,  2048);	/*  1/32 gray */
  SetColorEntry(17,  4096,  4096,  4096);	/*  2/32 gray */
  SetColorEntry(18,  6144,  6144,  6144);	/*  3/32 gray */
  SetColorEntry(19, 10240, 10240, 10240);	/*  5/32 gray */
  SetColorEntry(20, 12288, 12288, 12288);	/*  6/32 gray */
  SetColorEntry(21, 14336, 14336, 14336);	/*  7/32 gray */
  SetColorEntry(22, 18432, 18432, 18432);	/*  9/32 gray */
  SetColorEntry(23, 20480, 20480, 20480);	/* 10/32 gray */
  SetColorEntry(24, 22528, 22528, 22528);	/* 11/32 gray */
  SetColorEntry(25, 26624, 26624, 26624);	/* 13/32 gray */
  SetColorEntry(26, 28672, 28672, 28672);	/* 14/32 gray */
  SetColorEntry(27, 30720, 30720, 30720);	/* 15/32 gray */
  SetColorEntry(28, 34815, 34815, 34815);	/* 17/32 gray */
  SetColorEntry(29, 36863, 36863, 36863);	/* 18/32 gray */
  SetColorEntry(30, 38911, 38911, 38911);	/* 19/32 gray */
  SetColorEntry(31, 43007, 43007, 43007);	/* 21/32 gray */
  SetColorEntry(32, 45055, 45055, 45055);	/* 22/32 gray */
  SetColorEntry(33, 47103, 47103, 47103);	/* 23/32 gray */
  SetColorEntry(34, 51199, 51199, 51199);	/* 25/32 gray */
  SetColorEntry(35, 53247, 53247, 53247);	/* 26/32 gray */
  SetColorEntry(36, 55295, 55295, 55295);	/* 27/32 gray */
  SetColorEntry(37, 59391, 59391, 59391);	/* 29/32 gray */
  SetColorEntry(38, 61439, 61439, 61439);	/* 30/32 gray */
  SetColorEntry(39, 63487, 63487, 63487);	/* 31/32 gray */

  /* The remainder of color table defines a color cube with six steps
     for each primary color. Note that the corners of this cube repeat
     previous colors, but simplifies the mapping between RGB colors and
     color map indices. This color cube spans indices 40 through 255.
     */
  {
    int r, g, b;

    for (r= 0; r < 6; r++)
      for (g= 0; g < 6; g++)
	for (b= 0; b < 6; b++)
	  {
	    int i= 40 + ((36 * r) + (6 * b) + g);
	    if (i > 255) error("index out of range in color table compuation");
	    SetColorEntry(i, (r * 65535) / 5, (g * 65535) / 5, (b * 65535) / 5);
	  }
  }

  palette = CreatePalette(logPal);


  /* generate BITMAPINFOHEADERs */
  /* 1 bit color depth */
  bmi1 = (BITMAPINFO*) calloc(1,sizeof(BITMAPINFO) + 2 * sizeof(RGBQUAD));
  bmi1->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi1->bmiHeader.biPlanes = 1;
  bmi1->bmiHeader.biBitCount = 1;
  bmi1->bmiHeader.biCompression = BI_RGB;
  for(i=0;i<2;i++)
    {
      bmi1->bmiColors[i].rgbRed = logPal->palPalEntry[i].peRed;
      bmi1->bmiColors[i].rgbGreen = logPal->palPalEntry[i].peGreen;
      bmi1->bmiColors[i].rgbBlue = logPal->palPalEntry[i].peBlue;
    }
  /* 4 bit color depth */
  bmi4 = (BITMAPINFO*) calloc(1,sizeof(BITMAPINFO) + 16 * sizeof(RGBQUAD));
  bmi4->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi4->bmiHeader.biPlanes = 1;
  bmi4->bmiHeader.biBitCount = 4;
  bmi4->bmiHeader.biCompression = BI_RGB;
  for(i=0;i<16;i++)
    {
      bmi4->bmiColors[i].rgbRed = logPal->palPalEntry[i].peRed;
      bmi4->bmiColors[i].rgbGreen = logPal->palPalEntry[i].peGreen;
      bmi4->bmiColors[i].rgbBlue = logPal->palPalEntry[i].peBlue;
    }
  /* 8 bit color depth */
  bmi8 = (BITMAPINFO*) calloc(1,sizeof(BITMAPINFO) + 256 * sizeof(RGBQUAD));
  bmi8->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi8->bmiHeader.biPlanes = 1;
  bmi8->bmiHeader.biBitCount = 8;
  bmi8->bmiHeader.biCompression = BI_RGB;
  for(i=0;i<256;i++)
    {
      bmi8->bmiColors[i].rgbRed = logPal->palPalEntry[i].peRed;
      bmi8->bmiColors[i].rgbGreen = logPal->palPalEntry[i].peGreen;
      bmi8->bmiColors[i].rgbBlue = logPal->palPalEntry[i].peBlue;
    }
  /* 16 bit color depth */
  bmi16 = (BITMAPINFO*) calloc(1,sizeof(BITMAPINFO) + 4 * sizeof(DWORD));
  bmi16->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi16->bmiHeader.biPlanes = 1;
  bmi16->bmiHeader.biBitCount = 16;
  bmi16->bmiHeader.biCompression = BI_BITFIELDS;
  ((DWORD*) bmi16->bmiColors)[0] = (31 << 10); /* red mask */
  ((DWORD*) bmi16->bmiColors)[1] = (31 << 5); /* green mask */
  ((DWORD*) bmi16->bmiColors)[2] = (31 << 0); /* blue mask */

  /* 32 bit color depth */
  bmi32 = (BITMAPINFO*) calloc(1,sizeof(BITMAPINFO) + 4 * sizeof(DWORD));
  bmi32->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bmi32->bmiHeader.biPlanes = 1;
  bmi32->bmiHeader.biBitCount = 32;
  bmi32->bmiHeader.biCompression = BI_BITFIELDS;
  ((DWORD*) bmi32->bmiColors)[0] = 0x00FF0000; /* red mask */
  ((DWORD*) bmi32->bmiColors)[1] = 0x0000FF00; /* green mask */
  ((DWORD*) bmi32->bmiColors)[2] = 0x000000FF; /* blue mask */

}

#endif /* NO_STANDARD_COLORS */

/****************************************************************************/
/*                     Window Setup                                         */
/****************************************************************************/

/* SetWindowTitle(): Set the main window title */
void SetWindowTitle()
{ 
  TCHAR titleString[MAX_PATH+20];

  if(!IsWindow(stWindow)) return;
  wsprintf(titleString,TEXT("Squeak! (%s)"), toUnicode(imageName));
  SetWindowText(stWindow,titleString);
}

void SetupWindows()
{ WNDCLASS wc;

  /* create our update region */
  updateRgn = CreateRectRgn(0,0,1,1);

  /* No windows at all when running as NT service */
  if(fRunService && !fWindows95) return;

  wc.style = CS_OWNDC; /* don't waste resources ;-) */
  wc.lpfnWndProc = (WNDPROC)MainWndProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon (hInstance, MAKEINTRESOURCE(2));
  wc.hCursor = NULL;
  wc.hbrBackground = (HBRUSH)GetStockObject (BLACK_BRUSH);
  wc.lpszMenuName = NULL;
  wc.lpszClassName = TEXT("SqueakWindowClass");

  /* Due to Win95 bug we have to ignore the return value
     of RegisterClass(). Win95 returns several strange
     error messages (such as ERR_FUNCTION_NOT_SUPPORTED)
     when the class is already registered (which should
     result in an ERR_CLASS_ALREADY_EXISTS) */
  RegisterClass(&wc);

  if(!browserWindow)
    stWindow = CreateWindowEx(
#if !defined(_WIN32_WCE)
      			      WS_EX_APPWINDOW |
#endif
    			      WS_EX_OVERLAPPEDWINDOW,
			      TEXT("SqueakWindowClass"),
			      TEXT("Squeak!"),
#if !defined(_WIN32_WCE)
			      WS_OVERLAPPEDWINDOW /* | WS_CLIPCHILDREN */,
#else
			      WS_BORDER | WS_VISIBLE,
#endif
			      0,
			      0,
			      CW_USEDEFAULT,
			      CW_USEDEFAULT,
			      NULL,
			      NULL,
			      hInstance,
			      NULL);
  else {
    /* Setup a browser window. */
    fBrowserMode = 1;
    stWindow = CreateWindowEx(0,
			      TEXT("SqueakWindowClass"),
			      TEXT("Squeak!"),
			      WS_CHILD /* | WS_CLIPCHILDREN */,
			      0,
			      0,
			      GetSystemMetrics(SM_CXSCREEN),
			      GetSystemMetrics(SM_CYSCREEN),
			      browserWindow,
			      NULL,
			      hInstance,
			      NULL);
  }

#ifndef NO_WHEEL_MOUSE
  g_WM_MOUSEWHEEL = RegisterWindowMessage( TEXT("MSWHEEL_ROLLMSG") ); /* RvL 1999-04-19 00:23 */
#endif

#if defined(_WIN32_WCE)

#if _WIN32_WCE >= 300 && !defined(USE_GAPI)
  // Create a MenuBar
  {
         SHMENUBARINFO mbi;
         memset(&mbi, 0, sizeof(SHMENUBARINFO));
         mbi.cbSize            = sizeof(SHMENUBARINFO);
         mbi.hwndParent        = stWindow;
         mbi.dwFlags           = SHCMBF_EMPTYBAR;
         if (!SHCreateMenuBar(&mbi))
                 MessageBox(stWindow, L"SHCreateMenuBar Failed", L"Error", MB_OK);
  }
#endif

  /* WinCE does not support RegisterClassEx(), so we must set 
     the small icon after creating the window. */
  SendMessage(stWindow,WM_SETICON, FALSE, 
	      (LPARAM)LoadImage(hInstance, MAKEINTRESOURCE(1), 
				IMAGE_ICON, 16, 16, 0));
  
  consoleWindow = NULL; /* We do not use console under WinCE */

#else /* defined(_WIN32_WCE) */

  consoleWindow = CreateWindowEx(0,
				 TEXT("EDIT"),
				 TEXT(""),
				 WS_CHILD | WS_BORDER | 
				 WS_HSCROLL | WS_VSCROLL | ES_MULTILINE,
				 0,
				 0,
				 CW_USEDEFAULT,
				 CW_USEDEFAULT,
				 stWindow,
				 NULL,
				 hInstance,
				 NULL);
#endif /* defined(_WIN32_WCE) */

  /* Modify the system menu for any VM options */
  CreatePrefsMenu();
  SetWindowTitle();
  SetForegroundWindow(stWindow);
#ifndef NO_DROP
  /* drag and drop needs to be set up on per-window basis */
  SetupDragAndDrop();
#endif
  ioScreenSize(); /* compute new rect initially */
}


#if !defined(_WIN32_WCE)  /* Unused under WinCE */

void SetWindowSize(void) 
{ RECT r;
  int width, height, maxWidth, maxHeight, actualWidth, actualHeight;

  if(!IsWindow(stWindow)) return; /* might happen if run as NT service */
  if(browserWindow) return; /* Ignored if in browser */

  if (savedWindowSize != 0) 
    {
      width  = (unsigned) savedWindowSize >> 16;
      height = savedWindowSize & 0xFFFF;
    } 
  else 
    {
      width  = 640;
      height = 480;
    }

  /* minimum size is 64 x 64 */
  width  = ( width > 64) ?   width : 64;
  height = (height > 64) ?  height : 64;

  /* maximum size is screen size */
  maxWidth  = GetSystemMetrics(SM_CXFULLSCREEN);
  maxHeight = GetSystemMetrics(SM_CYFULLSCREEN);
  width  = ( width <= maxWidth)  ?  width : maxWidth;
  height = (height <= maxHeight) ? height : maxHeight;

  SetWindowPos(stWindow, 
  		NULL, 
  		(maxWidth-width) / 2, 
  		(maxHeight-height) / 2, 
  		width, 
  		height, 
  		SWP_NOZORDER | SWP_HIDEWINDOW);

  /* Get the client area to recompute the window size accordingly */
  GetClientRect(stWindow,&r);
  actualWidth = r.right - r.left;
  actualHeight = r.bottom - r.top;
  width += (width - actualWidth);
  height += (height - actualHeight);
  width  = ( width <= maxWidth)  ?  width : maxWidth;
  height = (height <= maxHeight) ? height : maxHeight;

  SetWindowPos(stWindow, 
  		NULL, 
  		(maxWidth-width) / 2, 
  		(maxHeight-height) / 2, 
  		width, 
  		height, 
  		SWP_NOZORDER | SWP_HIDEWINDOW);
}

#endif /* !defined(_WIN32_WCE) */

/****************************************************************************/
/*              Keyboard and Mouse                                          */
/****************************************************************************/

/* The following is a mapping to Mac Roman glyphs.
   It is not entirely correct since a number of glyphs are
   different but should be good enough for Squeak.
   More significantly, we can now map in both directions. */
static unsigned char keymap[256] =
{
  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79,
 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
 96, 97, 98, 99,100,101,102,103,104,105,106,107,108,109,110,111,
112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
173,176,226,196,227,201,160,224,246,228,178,220,206,179,182,183,
184,212,213,210,211,165,208,209,247,170,185,221,207,186,189,217,
202,193,162,163,219,180,195,164,172,169,187,199,194,197,168,248,
161,177,198,215,171,181,166,225,252,218,188,200,222,223,240,192,
203,231,229,204,128,129,174,130,233,131,230,232,237,234,235,236,
245,132,241,238,239,205,133,249,175,244,242,243,134,250,251,167,
136,135,137,139,138,140,190,141,143,142,144,145,147,146,148,149,
253,150,152,151,153,155,154,214,191,157,156,158,159,254,255,216
};

/* The following is the inverse keymap */
static unsigned char iKeymap[256];

void SetupKeymap()
{ int i;
  for(i=0;i<256;i++)
    iKeymap[keymap[i]] = i;
}


/* Map a virtual key into something the Mac understands */
int mapVirtualKey(int virtKey)
{
  switch (virtKey) {
#ifdef MAP_HARDWARE_KEY
    case 193:       return 188;
    case 194:       return 255;
    case 195:       return 229;
    case 196:       return 230;
    case 197:       return 239;
#endif
    case VK_DELETE: return 127;
    case VK_INSERT: return 5;
    case VK_PRIOR:  return 11;
    case VK_NEXT :  return 12;
    case VK_END  :  return 4;
    case VK_HOME :  return 1;
    case VK_LEFT :  return 28;
    case VK_RIGHT:  return 29;
    case VK_UP   :  return 30;
    case VK_DOWN :  return 31;
    case VK_RETURN: return 13;
  }
  return 0;
}

/****************************************************************************/
/*              Event based primitive set                                   */
/****************************************************************************/
#define MAX_EVENT_BUFFER 500
static sqInputEvent eventBuffer[MAX_EVENT_BUFFER];
static int eventBufferGet = 0;
static int eventBufferPut = 0;

static sqInputEvent *nextEventPut(void) {
  sqInputEvent *evt;
  evt = eventBuffer + eventBufferPut;
  eventBufferPut = (eventBufferPut + 1) % MAX_EVENT_BUFFER;
  if (eventBufferGet == eventBufferPut) {
    /* buffer overflow; drop the last event */
    eventBufferGet = (eventBufferGet + 1) % MAX_EVENT_BUFFER;
  }
  return evt;
}

int recordMouseEvent(MSG *msg) {
  sqMouseEvent *evt;
  int alt, shift, ctrl, red, blue, yellow;
  if(!msg) return 0;

  evt = (sqMouseEvent*) nextEventPut();

  alt = GetKeyState(VK_MENU) & 0x8000;
  shift = msg->wParam & MK_SHIFT;
  ctrl = msg->wParam & MK_CONTROL;
  red = msg->wParam & MK_LBUTTON;
  if(f1ButtonMouse) {
    /* there's just a single button y'know */
    red |= msg->wParam & MK_MBUTTON;
    red |= msg->wParam & MK_RBUTTON;
    blue = yellow = 0;
  } else if(f3ButtonMouse) {
    blue   = msg->wParam & MK_MBUTTON;
    yellow = msg->wParam & MK_RBUTTON;
  } else {
    blue   = msg->wParam & MK_RBUTTON;
    yellow = msg->wParam & MK_MBUTTON;
  }
  if(red && !blue && !yellow) { /* red button honors modifiers */
    if(alt && !ctrl) {
      red = 0; yellow = 1;
    }
    if(ctrl && !alt) {
      red = 0; blue = 1;
    }
  }
  /* first the basics */
  evt->type = EventTypeMouse;
  evt->timeStamp = msg->time;

#ifndef LANDSCAPE_SCREEN
  evt->x = (int)(short)LOWORD(msg->lParam);
  evt->y = (int)(short)HIWORD(msg->lParam);
#else
  evt->x = g_gxdp.cyHeight - ((int)(short)HIWORD(msg->lParam));
  evt->y = (int)(short)LOWORD(msg->lParam);
#endif
  /* then the buttons */
  evt->buttons = 0;
  evt->buttons |= red ? RedButtonBit : 0;
  evt->buttons |= blue ? BlueButtonBit : 0;
  evt->buttons |= yellow ? YellowButtonBit : 0;
  /* then the modifiers */
  evt->modifiers = 0;
  evt->modifiers |= shift ? ShiftKeyBit : 0;
  evt->modifiers |= ctrl ? CtrlKeyBit : 0;
  evt->modifiers |= alt ? CommandKeyBit : 0;
  /* clean up reserved */
  evt->reserved1 = 0;
  evt->reserved2 = 0;
  return 1;
}

int recordDragDropEvent(HWND wnd, int dragType, int x, int y, int numFiles)
{
  sqDragDropFilesEvent *evt;
  int alt, shift, ctrl, modifiers;

  evt = (sqDragDropFilesEvent*) nextEventPut();

  alt = GetKeyState(VK_MENU) & 0x8000;
  shift = (GetKeyState(VK_SHIFT) & 0x8000);
  ctrl = GetKeyState(VK_CONTROL) & 0x8000;

  /* first the basics */
  evt->type = EventTypeDragDropFiles;
  evt->timeStamp = ioMicroMSecs();
  evt->dragType = dragType;
  evt->x = x;
  evt->y = y;
  /* then the modifiers */
  modifiers = 0;
  modifiers |= shift ? ShiftKeyBit : 0;
  modifiers |= ctrl ? CtrlKeyBit : 0;
  modifiers |= alt ? CommandKeyBit : 0;
  evt->modifiers = modifiers;
  evt->numFiles = numFiles;

  /* clean up reserved */
  evt->reserved1 = 0;
  return 1;
}

int recordKeyboardEvent(MSG *msg) {
  sqKeyboardEvent *evt;
  int alt, shift, ctrl;
  int keyCode, virtCode, pressCode;

  if(!msg) return 0;

  alt = GetKeyState(VK_MENU) & 0x8000;
  shift = (GetKeyState(VK_SHIFT) & 0x8000);
  ctrl = GetKeyState(VK_CONTROL) & 0x8000;
  /* now the key code */
  virtCode = mapVirtualKey(msg->wParam);
  keyCode = keymap[msg->wParam];
  /* press code must differentiate */
  switch(msg->message) {
    case WM_KEYDOWN:
    case WM_SYSKEYDOWN:
      if(virtCode) keyCode = virtCode;
      pressCode = EventKeyDown;
      break;
    case WM_KEYUP:
    case WM_SYSKEYUP: 
      pressCode = EventKeyUp; 
      break;
    case WM_CHAR:
    case WM_SYSCHAR:
      /* Note: VK_RETURN is recorded as virtual key ONLY */
      if(keyCode == 13) return 1;
      pressCode = EventKeyChar;
      break;
    default: 
      pressCode = EventKeyChar;
  }
  /* remove Ctrl+Alt codes for international keyboards */
  if(ctrl && alt) {
    ctrl = 0;
    alt = 0;
  }
  /* first the basics */
  evt = (sqKeyboardEvent*) nextEventPut();
  evt->type = EventTypeKeyboard;
  evt->timeStamp = msg->time;
  evt->charCode = keyCode;
  evt->pressCode = pressCode;
  evt->modifiers = 0;
  evt->modifiers |= alt ? CommandKeyBit : 0;
  evt->modifiers |= shift ? ShiftKeyBit : 0;
  evt->modifiers |= ctrl ? CtrlKeyBit : 0;
  /* clean up reserved */
  evt->reserved1 = 0;
  evt->reserved2 = 0;
  /* note: several keys are not reported as character events;
     most noticably the mapped virtual keys. For those we
     generate extra character events here */
  if(pressCode == EventKeyDown && virtCode != 0) {
    /* generate extra character event */
    sqKeyboardEvent *extra = (sqKeyboardEvent*)nextEventPut();
    *extra = *evt;
    extra->pressCode = EventKeyChar;
  }
  return 1;
}

int ioSetInputSemaphore(int semaIndex) {
  inputSemaphoreIndex = semaIndex;
  return 1;
}

int ioGetNextEvent(sqInputEvent *evt) {
  if (eventBufferGet == eventBufferPut) {
    ioProcessEvents();
  }
  if (eventBufferGet == eventBufferPut) 
    return 1;
  
  *evt = eventBuffer[eventBufferGet];
  eventBufferGet = (eventBufferGet+1) % MAX_EVENT_BUFFER;
  return 1;
}

/****************************************************************************/
/*              State based primitive set                                   */
/****************************************************************************/

int ioGetKeystroke(void)
{
  int keystate;
  ioProcessEvents();  /* process all pending events */
  if (keyBufGet == keyBufPut) return -1;  /* keystroke buffer is empty */
  keystate= keyBuf[keyBufGet];
  keyBufGet= (keyBufGet + 1) % KEYBUF_SIZE;
  /* set modifer bits in buttonState to reflect the last keystroke fetched */
  buttonState= ((keystate >> 5) & 0xF8) | (buttonState & 0x7);
  return keystate;
}

int ioPeekKeystroke(void)
{
  int keystate;
  ioProcessEvents();  /* process all pending events */
  if (keyBufGet == keyBufPut) return -1;  /* keystroke buffer is empty */
  keystate= keyBuf[keyBufGet];
  /* set modifer bits in buttonState to reflect the last keystroke peeked at */
  buttonState= ((keystate >> 5) & 0xF8) | (buttonState & 0x7);
  return keystate;
}


void recordKey(int keystate)
{
  keyBuf[keyBufPut]= keystate;
  keyBufPut= (keyBufPut + 1) % KEYBUF_SIZE;
  if (keyBufGet == keyBufPut) {
    /* buffer overflow; drop the last character */
    keyBufGet= (keyBufGet + 1) % KEYBUF_SIZE;
    keyBufOverflows++;
  }
}

int recordVirtualKey(UINT message, WPARAM wParam, LPARAM lParam)
{
  int keystate = 0;

  if(wParam == VK_F2) {
    TrackPrefsMenu();
    return 1;
  }
  if(wParam == VK_CANCEL) {
    interruptPending= true;
    interruptCheckCounter= 0;
    return 1;
  }
  keystate = mapVirtualKey(wParam);
  if(keystate == 0) return 0;
  if(keystate == interruptKeycode)
    {
      /* NOTE: Interrupt key is meta, not recorded as key stroke */
      interruptPending= true;
      interruptCheckCounter= 0;
	  return 1;
    }
  keystate = keystate | ((buttonState >> 3) << 8);
  recordKey(keystate);
  return 1;
}

int recordKeystroke(UINT msg, WPARAM wParam, LPARAM lParam)
{
  int keystate=0;

  /* Special case: VK_RETURN is handled as virtual key *only* */
  if(wParam == 13) return 1;

  /* Map from Win32 to Mac */
  keystate = keymap[wParam];
  /* add the modifiers */
  keystate = keystate | ((buttonState >> 3) << 8);
  /* check for interrupt key */
  if(keystate == interruptKeycode)
    {
      /* NOTE: Interrupt key is meta, not recorded as key stroke */
      interruptPending= true;
      interruptCheckCounter= 0;
	  return 1;
    }
  recordKey(keystate);
  return 1;
}

/* record mouse events */
int recordMouseDown(WPARAM wParam, LPARAM lParam)
{
  int stButtons= 0;

#if defined(_WIN32_WCE)

  if (wParam & MK_LBUTTON) stButtons |= 4;
  if (stButtons == 4)	/* red button honours the modifiers */
    {				 
      if (wParam & MK_CONTROL) stButtons = 1;	/* blue button if CTRL down */
      else if (GetKeyState(VK_LMENU) & 0x8000) stButtons = 2;	/* yellow button if META down */
    }

#else /* defined(_WIN32_WCE) */

  if(GetKeyState(VK_LBUTTON) & 0x8000) stButtons |= 4;
  if(GetKeyState(VK_MBUTTON) & 0x8000) 
    if(f1ButtonMouse) stButtons |= 4;
    else stButtons |= f3ButtonMouse ? 2 : 1;
  if(GetKeyState(VK_RBUTTON) & 0x8000) 
    if(f1ButtonMouse) stButtons |= 4;
    else stButtons |= f3ButtonMouse ? 1 : 2;

  if (stButtons == 4)	/* red button honours the modifiers */
    {		     
      if (GetKeyState(VK_CONTROL) & 0x8000)
        stButtons= 2;	/* blue button if CTRL down */
      else if (GetKeyState(VK_MENU) & 0x8000)
        stButtons= 1;	/* yellow button if META down */
    }

#endif /* defined(_WIN32_WCE) */

  buttonState = stButtons & 0x7;
  return 1;
}

/* record the modifier buttons */
int recordModifierButtons()
{ int modifiers=0;

  /* map both shift and caps lock to squeak shift bit */
  if(GetKeyState(VK_SHIFT) & 0x8000)
    modifiers |= 1 << 3;

  /* Map meta key to command.
     
     NOTE:
     Non US keyboards typically need the right menu key for special mappings.
     However, it seems that hitting this key is interpreted as Alt-Ctrl.
     We therefore ignore the control modifier if the right menu key 
     is pressed (hoping that nobody will use the right-menu / control combination)
     
     NOTE^2:
     Due to another Win95 bug, we can not ask for the left
     or right alt key directly (Win95 returns always 0 if
     asked for VK_LMENU or VK_RMENU). Therefore we're using
     the Ctrl-Alt combination for detecting the right menu key */

  /* if CTRL and not ALT pressed use control modifier */
  if((GetKeyState(VK_CONTROL) & 0x8000) && (GetKeyState(VK_MENU) & 0x8000) == 0)
    modifiers |= 1 << 4;

  /* if ALT and not CTRL pressed use command modifier */
  if((GetKeyState(VK_MENU) & 0x8000) && (GetKeyState(VK_CONTROL) & 0x8000) == 0)
    modifiers |= 1 << 6;
  
  /* if ALT and CTRL pressed use option modifier */
  if((GetKeyState(VK_MENU) & 0x8000) && (GetKeyState(VK_CONTROL) & 0x8000))
    modifiers |= 1 << 5;

  /* button state: low three bits are mouse buttons; next 4 bits are modifier bits */
  buttonState= modifiers | (buttonState & 0x7);
  return 1;
}

int ioGetButtonState(void)
{
  if(fReduceCPUUsage || (fReduceCPUInBackground && !fHasFocus)) { 
    MSG msg;
    /* Delay execution of squeak if
       - there is currently no button pressed
       - there is currently no mouse event in the input queue
       by at most 5 msecs. This will reduce cpu load during idle times.
    */
    if((buttonState & 0x7) == 0 &&
       !PeekMessage(&msg, stWindow, WM_MOUSEFIRST, WM_MOUSELAST, PM_NOREMOVE))
      MsgWaitForMultipleObjects(0,NULL,0,5, QS_MOUSE);
  }
  ioProcessEvents();  /* process all pending events */
  return buttonState;
}

int ioMousePoint(void)
{ 
  if(fReduceCPUUsage || (fReduceCPUInBackground && !fHasFocus)) { 
    MSG msg;
    /* Delay execution of squeak if
       - there is currently no button pressed
       - there is currently no mouse event in the input queue
       by at most 5 msecs. This will reduce cpu load during idle times.
    */
    if((buttonState & 0x7) == 0 &&
       !PeekMessage(&msg, stWindow, WM_MOUSEFIRST, WM_MOUSELAST, PM_NOREMOVE))
      MsgWaitForMultipleObjects(0,NULL,0,5, QS_MOUSE);
  }
  ioProcessEvents();  /* process all pending events */
  /* x is high 16 bits; y is low 16 bits */
#ifndef LANDSCAPE_SCREEN
  return (mousePosition.x << 16) | (mousePosition.y);
#else
  return ((g_gxdp.cyHeight-mousePosition.y) << 16) | (mousePosition.x);
#endif 
}

/****************************************************************************/
/*              Misc support primitves                                      */
/****************************************************************************/

int ioExit(void)
{
#ifdef USE_BACKINGSTORE
  deleteBackingStoreFile();
#endif

  exit(0);
  /* avoid the warnings here */
  return 0;
}

int ioBeep(void)
{
  MessageBeep(0);
  return 1;
}

int ioMSecs()
{
  /* Make sure the value fits into Squeak SmallIntegers */
  return GetTickCount() &0x3FFFFFFF;
}

/* Note: ioMicroMSecs returns *milli*seconds */
int ioMicroMSecs(void)
{
  /* Make sure the value fits into Squeak SmallIntegers */
  return timeGetTime() &0x3FFFFFFF;
}

/* Note: ioRelinquishProcessorForMicroseconds has *micro*seconds  as argument*/
int ioRelinquishProcessorForMicroseconds(int microSeconds) 
{
  /* wake us up if something happens */
#ifndef POLL_NETWORK
  ResetEvent(vmWakeUpEvent);
  MsgWaitForMultipleObjects(1, &vmWakeUpEvent, FALSE, 
			    microSeconds / 1000, QS_ALLINPUT);
  return microSeconds;
#else
  return 0;
#endif
}

int ioProcessEvents(void)
{ static MSG msg;
  POINT mousePt;

#ifdef POLL_NETWORK
  sqNetPollSync();
#endif

  if(fRunService && !fWindows95) return 1;
  /* WinCE doesn't retrieve WM_PAINTs from the queue with PeekMessage,
     so we won't get anything painted unless we use GetMessage() if there
     is a dirty rect. */
  lastMessage = &msg;
  while(PeekMessage(&msg,NULL,0,0,PM_NOREMOVE))
    {
      GetMessage(&msg,NULL,0,0);
#ifndef NO_PLUGIN_SUPPORT
      if(msg.hwnd == NULL) {
	pluginHandleEvent(&msg);
      } else
#endif
	if(msg.hwnd != stWindow) {
	  /* Messages not sent to Squeak window */
	  if(msg.hwnd != consoleWindow && GetParent(msg.hwnd) == stWindow) {
	    /* This message has been sent to a plugin window */
	    /* Selectively pass up certain events to the parent's level */
	    switch (msg.message) {
	      case WM_LBUTTONDOWN:
	      case WM_LBUTTONUP:
	      case WM_MBUTTONDOWN:
	      case WM_MBUTTONUP:
	      case WM_RBUTTONDOWN:
	      case WM_RBUTTONUP:
	      case WM_MOUSEMOVE:
		mousePt.x = LOWORD(msg.lParam);
		mousePt.y = HIWORD(msg.lParam);
		MapWindowPoints(msg.hwnd, stWindow, &mousePt, 1);
		PostMessage(stWindow, msg.message, msg.wParam, MAKELONG(mousePt.x,mousePt.y));
	    }
	  }
	}
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }

  /* If we're running in a browser check if the browser's still there */
  if(fBrowserMode && browserWindow) {
    if(!IsWindow(browserWindow)) ioExit();
  }
  lastMessage = NULL;
  return 1;
}

/* returns the size of the Squeak window */
int ioScreenSize(void)
{ 
  static RECT r;

  if(!IsWindow(stWindow)) return savedWindowSize;

#ifndef _WIN32_WCE
  if(browserWindow && GetParent(stWindow) == browserWindow) {
    GetClientRect(browserWindow,&r);
  } else {
    if(!IsIconic(stWindow))
      GetClientRect(stWindow,&r);
  }
  /* width is high 16 bits; height is low 16 bits */
  return MAKELONG(r.bottom,r.right);
#else /* _WIN32_WCE */
  if (fullScreenFlag) {
# ifdef USE_GAPI
#  ifndef LANDSCAPE_SCREEN
    return MAKELONG(g_gxdp.cyHeight, g_gxdp.cxWidth);
#  else
    return MAKELONG(g_gxdp.cxWidth, g_gxdp.cyHeight);
# endif
# else
    GetClientRect(stWindow,&r);
# endif
  } else {
    GetClientRect(stWindow,&r);
  }
  /* width is high 16 bits; height is low 16 bits */
# ifndef LANDSCAPE_SCREEN
  return MAKELONG(r.bottom,r.right);
# else
  return MAKELONG(r.right, r.bottom);
# endif
#endif
}

/* returns the depth of the OS display */
int ioScreenDepth(void) {
  int depth;
  HDC dc = GetDC(stWindow);
  if(!dc) return 0; /* fail */
  depth = GetDeviceCaps(dc, BITSPIXEL);
  ReleaseDC(stWindow, dc);
  return depth;
}

/* returns the local wall clock time */
int ioSeconds(void)
{ SYSTEMTIME sysTime;
  GetLocalTime(&sysTime);
  return convertToSqueakTime(sysTime);
}

int ioSetCursorWithMask(int cursorBitsIndex, int cursorMaskIndex, int offsetX, int offsetY)
{
#if !defined(_WIN32_WCE)
	/****************************************************/
	/* Only one cursor is defined under CE...           */
	/* (the wait cursor)                         :-(    */
	/****************************************************/
  static unsigned char *andMask=0,*xorMask=0;
  static int cx=0,cy=0,cursorSize=0;
  int i;

  if(!IsWindow(stWindow)) return 1;
  if(!andMask || !xorMask)
	{ /* Determine the cursor size.
		 NOTE: On NT this is actually not necessary, as NT _can_
		 create cursors of different sizes. However, on Win95 the
		 cursor will be created, BUT WILL NOT BE SHOWN! */
      cx = GetSystemMetrics(SM_CXCURSOR);
      cy = GetSystemMetrics(SM_CYCURSOR);
      cursorSize = cx*cy / 8;
      andMask = malloc(cursorSize);
      xorMask = malloc(cursorSize);
  }

  /* free last used cursor */
  if(currentCursor) DestroyCursor(currentCursor);

  memset(andMask,0xff,cursorSize);
  memset(xorMask,0x00,cursorSize);
  if(cursorMaskIndex)
    { /* New Cursors specify mask and bits:
         Mask    Bit    Result
         0       0      Transparent
         0       1      Invert
         1       0      White
         1       1      Black
      */
      for (i=0; i<16; i++)
        {
          andMask[i*cx/8+0] = ~(checkedLongAt(cursorMaskIndex + (4 * i)) >> 24) & 0xFF;
          andMask[i*cx/8+1] = ~(checkedLongAt(cursorMaskIndex + (4 * i)) >> 16) & 0xFF;
        }
      for (i=0; i<16; i++)
        {
          xorMask[i*cx/8+0] = (~(checkedLongAt(cursorBitsIndex + (4 * i)) >> 24) & 0xFF) ^ (andMask[i*cx/8+0]);
          xorMask[i*cx/8+1] = (~(checkedLongAt(cursorBitsIndex + (4 * i)) >> 16) & 0xFF) ^ (andMask[i*cx/8+1]);
        }
    }
  else /* Old Cursor: Just make all 1-bits black */
    for (i=0; i<16; i++)
      {
        andMask[i*cx/8+0] = ~(checkedLongAt(cursorBitsIndex + (4 * i)) >> 24) & 0xFF;
        andMask[i*cx/8+1] = ~(checkedLongAt(cursorBitsIndex + (4 * i)) >> 16) & 0xFF;
      }

  currentCursor = CreateCursor(hInstance,-offsetX,-offsetY,cx,cy,andMask,xorMask);
  if(currentCursor) 
    {
      SetCursor(0);
      SetCursor(currentCursor);
    }
  else
    {
      printLastError(TEXT("CreateCursor failed"));
    }
#endif /* !defined(_WIN32_WCE) */
  return 1;
}

int ioSetCursor(int cursorBitsIndex, int offsetX, int offsetY)
{
  return ioSetCursorWithMask(cursorBitsIndex, 0, offsetX, offsetY);
}


int ioSetFullScreen(int fullScreen)
{ static int wasFullScreen = 0;

  if(!IsWindow(stWindow)) return 1;
  if(wasFullScreen == fullScreen) return 1;
  /* NOTE: No modifications if the window is not currently
           visible, else we'll have a number of strange effects ... */
  if(!IsWindowVisible(stWindow)) return 1; 
  if(fullScreen)
    {
#if !defined(_WIN32_WCE)
      if(browserWindow) {
	/* Jump out of the browser */
	HWND oldBrowserWindow = browserWindow;
	browserWindow = NULL;
	DestroyWindow(stWindow);
	SetupWindows();
	/* Hm ... some weird problem on Win98 in IE doesn't
	   bring Squeak to top the way it's supposed to be.
	   So we do it the hard way ;-) */
	SetWindowPos(stWindow, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_SHOWWINDOW);
	browserWindow = oldBrowserWindow;
      }
      SetWindowLong(stWindow,GWL_STYLE, WS_POPUP | WS_CLIPCHILDREN);
      SetWindowLong(stWindow,GWL_EXSTYLE, WS_EX_APPWINDOW);
      ShowWindow(stWindow, SW_SHOWMAXIMIZED);
#else /* !defined(_WIN32_WCE) */
#  ifdef USE_GAPI
      ioCESetFullScreen(fullScreen);
#  else
      ShowWindow(stWindow,SW_SHOWNORMAL);
#  endif
#endif /* !defined(_WIN32_WCE) */
      fullScreenFlag = 1;
    }
  else
    {
#if !defined(_WIN32_WCE)
      ShowWindow(stWindow, SW_RESTORE);
      ShowWindow(stWindow, SW_HIDE);
      SetWindowLong(stWindow,GWL_STYLE, WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN);
      SetWindowLong(stWindow,GWL_EXSTYLE, WS_EX_APPWINDOW | WS_EX_OVERLAPPEDWINDOW);
      SetWindowPos(stWindow,0,0,0,0,0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER | SWP_FRAMECHANGED | SWP_NOREDRAW);
      if(browserWindow) {
	/* Jump back into the browser */
	DestroyWindow(stWindow);
	SetupWindows();
      }
#else
#  ifdef USE_GAPI
      ioCESetFullScreen(fullScreen);
#  endif
#endif /* !defined(_WIN32_WCE) */
      ShowWindow(stWindow,SW_SHOWNORMAL);
      fullScreenFlag = 0;
    }
  /* get us back in the foreground */
  SetForegroundWindow(stWindow);
  wasFullScreen = fullScreen;
  return 1;
}

/****************************************************************************/
/*                           Image byte reversal                            */
/****************************************************************************/
#ifdef MSB_FIRST
# define BYTE_SWAP(w) w
# define WORD_SWAP(w) w
#else /* LSB_FIRST */
# if defined(__GNUC__) && defined(_X86_)
   /* GCC generates *optimal* code with a little help */
#  define BYTE_SWAP(w) __asm__("bswap %%eax" : "=eax" (w) : "eax" (w))
#  define WORD_SWAP(w) __asm__("roll $16, %%eax" : "=eax" (w) : "eax" (w))
#  define SRC_PIX_REG asm("%esi")
#  define DST_PIX_REG asm("%edi")
# else /* Not GCC?! Well, it's your own fault */
#  define BYTE_SWAP(w) w = (w << 24) | ((w & 0xFF00) << 8) | ((w >> 8) & 0xFF00) | (w >> 24)
#  define WORD_SWAP(w) w = (( (unsigned)(w) << 16) | ((unsigned) (w) >> 16))
# endif /* __GNUC__ */
#endif /* MSB_FIRST */

#ifndef SRC_PIX_REG
# define SRC_PIX_REG
# define DST_PIX_REG
#endif

int reverse_image_bytes(unsigned int* dst, unsigned int *src,
			int depth, int width, RECT *rect)
{ 
  int pitch, first, last, nWords, delta, yy;

  /* --- SPECIAL HACK FOR WINDOWS CE --- */
#ifdef _WIN32_WCE
  int reverseBits = 0;
  if(depth == 1)
    reverseBits = 1;
#else
  /* compiler will optimize it away */
  static const int reverseBits = 0;
#endif

  /* note: all  of the below are in DWORDs not BYTEs */
  pitch = ((width * depth) + 31) / 32;
  first = (rect->left * depth) / 32;
  last  = ((rect->right * depth) + 31) / 32;
  nWords = last - first;
  delta = pitch - nWords;
  if(nWords <= 0) return 1;

  { /* the inner loop */
    register DWORD* srcPixPtr SRC_PIX_REG;
    register DWORD* dstPixPtr DST_PIX_REG;
    srcPixPtr = ((DWORD*)src) + (rect->top * pitch) + first;
    dstPixPtr = ((DWORD*)dst) + (rect->top * pitch) + first;
    if(reverseBits) {
      for(yy = rect->top; yy < rect->bottom; 
	  yy++, srcPixPtr += delta, dstPixPtr += delta) {
	int i = nWords;
	do {
	  unsigned int value = *srcPixPtr++;
	  BYTE_SWAP(value);
	  *dstPixPtr++ = ~value;
	} while(--i);
      }
    } else { /* !reverseBits */
      for(yy = rect->top; yy < rect->bottom; 
	  yy++, srcPixPtr += delta, dstPixPtr += delta) {
	int i = nWords;
	do {
	  unsigned int value = *srcPixPtr++;
	  BYTE_SWAP(value);
	  *dstPixPtr++ = value;
	} while(--i);
      }
    }
  }
  return 1;
}

int reverse_image_words(unsigned int *dst, unsigned int *src,
			int depth, int width, RECT *rect)
{ 
  int pitch, first, last, nWords, delta, yy;

  /* note: all  of the below are in DWORDs not BYTEs */
  pitch = ((width * depth) + 31) / 32;
  first = (rect->left * depth) / 32;
  last  = ((rect->right * depth) + 31) / 32;
  nWords = last - first;
  delta = pitch - nWords;
  if(nWords <= 0) return 1;

  { /* the inner loop */
    register DWORD* srcPixPtr SRC_PIX_REG;
    register DWORD* dstPixPtr DST_PIX_REG;
    srcPixPtr = ((DWORD*)src) + (rect->top * pitch) + first;
    dstPixPtr = ((DWORD*)dst) + (rect->top * pitch) + first;
    for(yy = rect->top; yy < rect->bottom; 
	yy++, srcPixPtr += delta, dstPixPtr += delta) {
      int i = nWords;
      do {
	unsigned int value = *srcPixPtr++;
	WORD_SWAP(value);
	*dstPixPtr++ = value;
      } while(--i);
    }
  }
  return 1;
}

int copy_image_words(unsigned int *dst, unsigned int *src,
		     int depth, int width, RECT *rect)
{ 
  int pitch, first, last, nWords, delta, yy;

  /* note: all  of the below are in DWORDs not BYTEs */
  pitch = ((width * depth) + 31) / 32;
  first = (rect->left * depth) / 32;
  last  = ((rect->right * depth) + 31) / 32;
  nWords = last - first;
  delta = pitch - nWords;
  if(nWords <= 0) return 1;

  {/* the inner loop */
    DWORD* srcPixPtr;
    DWORD* dstPixPtr;
    srcPixPtr = ((DWORD*)src) + (rect->top * pitch) + first;
    dstPixPtr = ((DWORD*)dst) + (rect->top * pitch) + first;
    for(yy = rect->top; yy < rect->bottom; 
	yy++, srcPixPtr += pitch, dstPixPtr += pitch) {
      memcpy(dstPixPtr, srcPixPtr, nWords*4);
    }
  }
  return 1;
}

#undef BYTE_SWAP
#undef WORD_SWAP
#undef SRC_PIX_REG
#undef DST_PIX_REG

/****************************************************************************/
/*              Display and printing                                        */
/****************************************************************************/

BITMAPINFO *BmiForDepth(int depth)
{ BITMAPINFO *bmi = NULL;
  switch(depth) {
    case 1: bmi = bmi1; break;
    case 4: bmi = bmi4; break;
    case 8: bmi = bmi8; break;
    case 16: bmi = bmi16; break;
    case 32: bmi = bmi32; break;
  }
  return bmi;
}

int ioHasDisplayDepth(int depth) {
  if(depth == 1 || depth == 4 || depth == 8 || depth == 16 || depth == 32)
    return 1;
  return 0;
}


int ioSetDisplayMode(int width, int height, int depth, int fullscreenFlag)
{
#ifdef _WIN32_WCE
  return 0; /* Not implemented on CE */
#else
  RECT r;
  static int wasFullscreen = 0;
  static HWND oldBrowserWindow = NULL;


  if(!IsWindow(stWindow)) return 0;
  if(!IsWindowVisible(stWindow)) return 0;

#ifdef USE_DIRECT_X
  if(wasFullscreen && !fullscreenFlag) {
    /* Some weird DirectX bug - if exclusive mode has been set for
       the main window you'll never get out of it. That's why we
       destroy and recreate the window here.
    */
    ioSetFullScreen(0); /* Turn off fullscreen */
    DirectXSetDisplayMode(stWindow, width, height, depth, 0);
    DestroyWindow(stWindow);
    browserWindow = oldBrowserWindow;
    SetupWindows();
    ShowWindow(stWindow, SW_SHOWNORMAL);
  }
  wasFullscreen = fullscreenFlag;
  if(fullscreenFlag && browserWindow) {
    /* Must get out of browser window */
    oldBrowserWindow = browserWindow;
    browserWindow = NULL;
    DestroyWindow(stWindow);
    SetupWindows();
    ShowWindow(stWindow, SW_SHOWNORMAL);
  }
  if(!DirectXSetDisplayMode(stWindow, width, height, depth, fullscreenFlag)) {
    /* We must carefully restore the old window here */
    if(oldBrowserWindow) {
      DestroyWindow(stWindow);
      browserWindow = oldBrowserWindow;
      SetupWindows();
      ShowWindow(stWindow, SW_SHOWNORMAL);
    }
    return 0;
  }
  /* Note: Only go to full screen if DirectX is used */
  ioSetFullScreen(fullscreenFlag);
#endif

  r.left = 0;
  r.top = 0;
  r.right = width;
  r.bottom = height;
  if(browserWindow) {
    r.right = GetSystemMetrics(SM_CXSCREEN);
    r.bottom = GetSystemMetrics(SM_CYSCREEN);
  } else {
    AdjustWindowRect(&r, GetWindowLong(stWindow, GWL_STYLE), 0);
  }
  SetWindowPos(stWindow, NULL, 0, 0, r.right-r.left, r.bottom-r.top,
	       SWP_NOMOVE | SWP_NOZORDER);
  SetFocus(stWindow);
  return 1;
#endif /* _WIN32_WCE */
}

/* force an update of the squeak window if using deferred updates */
int ioForceDisplayUpdate(void) {
  /* Check if
     a) We should do deferred updates at all
     b) The window is valid
     c) The Interpreter does not defer updates by itself
  */
  if(fDeferredUpdate && IsWindow(stWindow) && !deferDisplayUpdates)
    {
      UpdateWindow(stWindow);
    }
  return 1;
}

int ioFormPrint(int bitsAddr, int width, int height, int depth, double hDPI, double vDPI, int landscapeFlag)
	/* print a form with the given bitmap, width, height, and depth at
	   the given horizontal and vertical scales in the given orientation */
{
#ifdef NO_PRINTER
  warnPrintf(TEXT("This VM does not support printing.\n"));
  return success(false);
#else /* !defined(NO_PRINTER) */
  DEVMODE *dmPtr;
  BITMAPINFO *bmi;
  DOCINFO di;
  HDC dc;
  DEVNAMES *devNames;
  TCHAR *namePtr;
  RECT targetRect;
  int scWidth;
  int scHeight;

  devNames = GlobalLock(printValues.hDevNames);
  if(!devNames)
    {
      warnPrintf(TEXT("No printer configured\n"));
      return false;
    }
  dmPtr = GlobalLock(printValues.hDevMode);
  if(dmPtr)
    {
      dmPtr->dmOrientation = landscapeFlag ? DMORIENT_LANDSCAPE : DMORIENT_PORTRAIT;
      dmPtr->dmFields |= DM_ORIENTATION;
    }
  namePtr = (TCHAR*) devNames;
  dc = CreateDC(namePtr+devNames->wDriverOffset, 
                namePtr+devNames->wDeviceOffset, 
                namePtr+devNames->wOutputOffset,
                dmPtr);
  GlobalUnlock(printValues.hDevMode);
  GlobalUnlock(printValues.hDevNames);

  if(!dc)
    {
      warnPrintf(TEXT("Unable to open printer.\n"));
      return false;
    }

  bmi = BmiForDepth(depth);
  if(!bmi) 
    {
      warnPrintf(TEXT("Color depth %d not supported"), depth);
      return false;
    }

  di.cbSize      = sizeof(DOCINFO);
  di.lpszDocName = TEXT("Squeak Print Job");
  di.lpszOutput  = NULL;

  StartDoc  (dc, &di);
  StartPage (dc);

  bmi->bmiHeader.biWidth = width;
  bmi->bmiHeader.biHeight = height;
  bmi->bmiHeader.biSizeImage = 0;

  /* set device coords to 0.001 inch per unit */

  SetMapMode(dc, MM_HIENGLISH);
  scWidth = (int)(width * 1000.0 / hDPI);
  scHeight = (int)(height * 1000.0 / vDPI );

  /* reverse the image bits if necessary */
  targetRect.left = targetRect.top = 0;
  targetRect.right = width;
  targetRect.bottom = height;
#ifndef NO_BYTE_REVERSAL
  if( depth < 32 )
    if(depth == 16)
      reverse_image_words((unsigned int*) bitsAddr, (unsigned int*) bitsAddr, 
			  depth, width, &targetRect);
    else
      reverse_image_bytes((unsigned int*) bitsAddr, (unsigned int*) bitsAddr, 
			  depth, width, &targetRect);
#endif /* NO_BYTE_REVERSAL */

  if(GDI_ERROR == StretchDIBits(dc,
                0,         /* dst_x */
                -scHeight, /* dst_y --- mapping mode changes positive y-axis */
                scWidth,   /* dst_w */
                scHeight,  /* dst_h */
                0,         /* src_x */
                0,         /* src_y */
                width,     /* src_w */
                height,    /* src_h */
                (void*) bitsAddr, /* bits */
                bmi,
                DIB_RGB_COLORS,
                SRCCOPY)) printLastError(TEXT("StretchDIBits failed"));

  /* reverse the image bits if necessary */
#ifndef NO_BYTE_REVERSAL
  if( depth < 32 )
    if(depth == 16)
      reverse_image_words((unsigned int*) bitsAddr, (unsigned int*) bitsAddr, 
			  depth, width, &targetRect);
    else
      reverse_image_bytes((unsigned int*) bitsAddr, (unsigned int*) bitsAddr, 
			  depth, width, &targetRect);
#endif /* NO_BYTE_REVERSAL */

  EndPage   (dc);
  EndDoc    (dc);
  DeleteDC  (dc);

	return true;
#endif /* NO_PRINTER */
}


#ifdef USE_DIB_SECTIONS

/* CreateBitmapDC():
   Create a device context for a DIB of the selected size.
*/
HDC CreateBitmapDC(HDC dc, int depth, int width, int height, void** pBitsOut)
{
  /* Cached DIBSection */
  static HBITMAP hbm = NULL;
  static int lastDepth = 0;
  static int lastWidth = 0;
  static int lastHeight = 0;
  static void* pBits;
  static HDC memDC;
  BITMAPINFO *bmi;
  
  bmi = BmiForDepth(depth);
  if(!bmi)
    abortMessage(TEXT("Fatal error: Color depth %d not supported"),depth);

  if (depth != lastDepth || width != lastWidth || height != lastHeight)
    {
      lastDepth = depth;
      lastHeight = height;
      lastWidth = width;
      bmi->bmiHeader.biWidth = width;
      bmi->bmiHeader.biHeight = -height;
      if (hbm)
	DeleteObject(hbm);
      hbm = CreateDIBSection(dc, bmi, DIB_RGB_COLORS, &pBits, NULL, 0);
      if (!hbm)
	abortMessage(TEXT("Fatal error: Cannot create device bitmap!"));
    }
  
  *pBitsOut = pBits;
  memDC = CreateCompatibleDC(dc);
  SelectObject(memDC, hbm);
  return memDC;
}

/* ReleaseBitmapDC():
   Clean up the given DC.
*/
void ReleaseBitmapDC(HDC memDC)
{
  DeleteDC(memDC);
}

#endif /* USE_DIB_SECTIONS */

#ifdef USE_GAPI

#  ifndef LANDSCAPE_SCREEN

void
iPaqFrameBufferShowDisplayBits1(int dispBitsIndex, int width, int height, int depth, 
				int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine1 = bytesPerLine(width, 1);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord1 = scanLine1 * affectedT + bytesPerLineRD(affectedL, 1);
  int lastWord1InLoop = scanLine1 * affectedT + bytesPerLineRD(affectedR, 1);
  int line;

  int aL = (affectedL>>5)<<5;
  int deltaX = g_gxdp.cbxPitch>>1;
  int deltaY = g_gxdp.cbyPitch;

  unsigned char *vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  int vramIndex = (g_gxdp.cbxPitch) * (aL + borderWidthX) + 
    (g_gxdp.cyHeight - 1 - affectedT - borderWidthY) * (g_gxdp.cBPP/8);

  int maxLine;

  if (fullScreenFlag) {
    maxLine = affectedB;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    maxLine = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top - BORDERWIDTH
                : (affectedB < info.rcSipRect.bottom ? affectedB : info.rcSipRect.bottom);
  }

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int*)(((char*)dispBitsIndex) + firstWord1);
    unsigned int *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;

    int m;
    unsigned int mask;
    unsigned int swapedFrom;
    mask = 0x80000000;
    swapedFrom = *from;
    for (m = 0; m < 32; m++) {
      if (toIndex >= affectedL && toIndex < affectedR) {
        *to = (mask & swapedFrom) ? 0 : 0xFFFF;
      }
      mask >>= 1;
      to += deltaX;
      toIndex++;
    }
    from++;

    limit = (unsigned int*)(((char*)dispBitsIndex) + lastWord1InLoop);
    while (from < limit) {
      mask = 0x80000000;
      swapedFrom = *from;
      for (m = 0; m < 32; m++) {
	*to = (mask & swapedFrom) ? 0 : 0xFFFF;
	mask >>= 1;
	to += deltaX;
      }
      toIndex += 32;
      from++;
    }
    mask = 0x80000000;
    swapedFrom = *from;
    for (m = 0; m < 32; m++) {
      if (toIndex >= affectedR) {
        break;
      }
      *to = (mask & swapedFrom) ? 0 : 0xFFFF;
      mask >>= 1;
      to += deltaX;
      toIndex++;
    }
    
    firstWord1 += scanLine1;
    lastWord1InLoop += scanLine1;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits4(int dispBitsIndex, int width, int height, int depth, 
				int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine4 = bytesPerLine(width, 4);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord4 = scanLine4 * affectedT + bytesPerLineRD(affectedL, 4);
  int lastWord4InLoop = scanLine4 * affectedT + bytesPerLineRD(affectedR, 4);
  int line;

  int aL = (affectedL>>3)<<3;
  int deltaX = g_gxdp.cbxPitch>>1;
  int deltaY = g_gxdp.cbyPitch;

  unsigned char *vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  int vramIndex = (g_gxdp.cbxPitch) * (aL + borderWidthX) + 
    (g_gxdp.cyHeight - 1 - affectedT - borderWidthY) * (g_gxdp.cBPP/8);

  int maxLine;

  if (fullScreenFlag) {
    maxLine = affectedB;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    maxLine = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top - BORDERWIDTH
                : (affectedB < info.rcSipRect.bottom ? affectedB : info.rcSipRect.bottom);
  }

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int*)(((unsigned char*)dispBitsIndex) + firstWord4);
    unsigned int *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;
    unsigned int pixels;

    int i;
    unsigned int mask;
    int shift;

    mask = 0xF0000000;
    shift = 28;
    pixels = *from;
    for (i = 0; i < 8; i++) {
      if (toIndex >= affectedL && toIndex < affectedR) {
	*to = stColors16[(pixels&mask)>>shift];
      }
      to += deltaX;
      toIndex++;
      mask >>= 4;
      shift -= 4;
    }
    from++;

    limit = (unsigned int*)(((unsigned char*)dispBitsIndex) + lastWord4InLoop);

    while (from < limit) {
      pixels = *from;
      mask = 0xF0000000;
      shift = 28;
      for (i = 0; i < 8; i++) {
        *to = stColors16[(pixels&mask)>>shift];
        to += deltaX;
        mask >>= 4;
        shift -= 4;
      }
      toIndex += 8;
      from++;
    }

    pixels = *from;
    mask = 0xF0000000;
    shift = 28;
    for (i = 0; i < 8; i++) {
      if (toIndex >= affectedR) {
	break;
      }
      *to = stColors16[(pixels&mask)>>shift];
      to += deltaX;
      toIndex++;
      mask >>= 4;
      shift -= 4;
    }
    
    firstWord4 += scanLine4;
    lastWord4InLoop += scanLine4;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits8(int dispBitsIndex, int width, int height, int depth, 
				int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine8 = bytesPerLine(width, 8);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord8 = scanLine8 * affectedT + bytesPerLineRD(affectedL, 8);
  int lastWord8InLoop = scanLine8 * affectedT + bytesPerLineRD(affectedR, 8);
  int line;

  int aL = (affectedL>>2)<<2;
  int deltaX = g_gxdp.cbxPitch>>1;
  int deltaY = g_gxdp.cbyPitch;

  unsigned char *vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  int vramIndex = (g_gxdp.cbxPitch) * (aL + borderWidthX) + 
    (g_gxdp.cyHeight - 1 - affectedT - borderWidthY) * (g_gxdp.cBPP/8);

  int maxLine;

  if (fullScreenFlag) {
    maxLine = affectedB;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    maxLine = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top - BORDERWIDTH
                : (affectedB < info.rcSipRect.bottom ? affectedB : info.rcSipRect.bottom);
  }

  for (line = affectedT; line < maxLine; line++) {
    unsigned char *from = ((unsigned char*)dispBitsIndex) + firstWord8;
    unsigned char *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);

    int toIndex = aL;
    int i;

    for (i = 3; i >= 0; i--) {
      if (toIndex >= affectedL && toIndex < affectedR) {
	*to = stColors16[from[i]];
      }
      to += deltaX;
      toIndex++;
    }
    from += 4;

    limit = ((unsigned char*)dispBitsIndex) + lastWord8InLoop;

    while (from < limit) {
      *to = stColors16[from[3]];
      to += deltaX;
      *to = stColors16[from[2]];
      to += deltaX;
      *to = stColors16[from[1]];
      to += deltaX;
      *to = stColors16[from[0]];
      to += deltaX;
      toIndex += 4;
      from += 4;
    }

    for (i = 3; i >= 0; i--) {
      if (toIndex >= affectedR) {
	break;
      }
      *to = stColors16[from[i]];
      to += deltaX;
      toIndex++;
    }
    
    firstWord8 += scanLine8;
    lastWord8InLoop += scanLine8;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits16(int dispBitsIndex, int width, int height, int depth, 
				 int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord16 = scanLine16 * affectedT + bytesPerLineRD(affectedL, 16);
  int lastWord16InLoop = scanLine16 * affectedT + bytesPerLineRD(affectedR, 16);
  int line;

  int aL = (affectedL>>1)<<1;
  int aR = ((affectedR+1)>>1)<<1;
  int deltaX = g_gxdp.cbxPitch>>1;
  int deltaY = g_gxdp.cbyPitch;

  unsigned char *vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  int vramIndex = (g_gxdp.cbxPitch) * (aL + borderWidthX) + 
    (g_gxdp.cyHeight - 1 - affectedT - borderWidthY) * (g_gxdp.cBPP/8);

  int maxLine;

  if (fullScreenFlag) {
    maxLine = affectedB;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    maxLine = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top - BORDERWIDTH
                : (affectedB < info.rcSipRect.bottom ? affectedB : info.rcSipRect.bottom);
  }

  for (line = affectedT; line < maxLine; line++) {
    unsigned short *from = (unsigned short *)(((unsigned char*)dispBitsIndex) + firstWord16);
    unsigned short *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;
    unsigned short pixel;

    if (aL != affectedL) {
      to += deltaX;
      pixel = *(from);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      to += deltaX;
      toIndex += 2;
      from += 2;
    }

    limit = (unsigned short *)(((unsigned char*)dispBitsIndex) + lastWord16InLoop);

    while (from < limit) {
      pixel = *(from+1);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      to += deltaX;
      pixel = *(from);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      to += deltaX;
      toIndex += 2;
      from += 2;
    }

    if (aR != affectedR) {
      pixel = *(from+1);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
    }

    firstWord16 += scanLine16;
    lastWord16InLoop += scanLine16;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits32(int dispBitsIndex, int width, int height, int depth, 
				 int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine32 = bytesPerLine(width, 32);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord32 = scanLine32 * affectedT + bytesPerLineRD(affectedL, 32);
  int lastWord32 = scanLine32 * affectedT + bytesPerLine(affectedR, 32);
  int line;

  int aL = affectedL;
  int deltaX = g_gxdp.cbxPitch>>1;
  int deltaY = g_gxdp.cbyPitch;

  unsigned char *vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);

  int vramIndex = (g_gxdp.cbxPitch) * (aL + borderWidthX) + 
    (g_gxdp.cyHeight - 1 - affectedT - borderWidthY) * (g_gxdp.cBPP/8);

  int maxLine;

  if (fullScreenFlag) {
    maxLine = affectedB;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    maxLine = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top - BORDERWIDTH
                : (affectedB < info.rcSipRect.bottom ? affectedB : info.rcSipRect.bottom);
  }

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int *)(((unsigned char*)dispBitsIndex) + firstWord32);
    unsigned int *limit = (unsigned int *)(((unsigned char*)dispBitsIndex) + lastWord32);
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    unsigned int pixel;
    while (from < limit) {
      /*
	0000 0000 1111 1111 0000 0000 0000 0000
	                    1111 1000 0000 0000
	0000 0000 0000 0000 1111 1111 0000 0000
	                    0000 0111 1110 0000
	0000 0000 0000 0000 0000 0000 1111 1111
	                    0000 0000 0001 1111
			    */
      pixel = *from;
      pixel =
	((pixel&0x00F80000)>>8)
	|((pixel&0x0000FC00)>>5)
	|((pixel&0x000000F8)>>3);
      *to = pixel;
      to += deltaX;
      from++;
    }
    firstWord32 += scanLine32;
    lastWord32 += scanLine32;
    vramIndex += deltaY;
  }
}

#  else /* LANDSCAPE_SCREEN */

void
iPaqFrameBufferShowDisplayBits1(int dispBitsIndex, int width, int height, int depth, 
				int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine1 = bytesPerLine(width, 1);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord1 = scanLine1 * affectedT + bytesPerLineRD(affectedL, 1);
  int lastWord1InLoop = scanLine1 * affectedT + bytesPerLineRD(affectedR, 1);
  int line;

  int aL = (affectedL>>5)<<5;
  int deltaX = 1;
  int deltaY = g_gxdp.cbxPitch;

  unsigned char *vramTopPtr;
  int vramIndex;

  int leftStart;
  int maxLine = affectedB;

  if (fullScreenFlag) {
    leftStart = 0;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    leftStart = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top
                : info.rcSipRect.bottom;
  }

  vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  vramIndex = (g_gxdp.cbxPitch) * affectedT + (aL + borderWidthY + leftStart) * (g_gxdp.cBPP/8);

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int*)(((char*)dispBitsIndex) + firstWord1);
    unsigned int *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;

    int m;
    unsigned int mask;
    unsigned int swapedFrom;
    mask = 0x80000000;
    swapedFrom = *from;
    for (m = 0; m < 32; m++) {
      if (toIndex >= affectedL && toIndex < affectedR) {
        *to = (mask & swapedFrom) ? 0 : 0xFFFF;
      }
      mask >>= 1;
      to += deltaX;
      toIndex++;
    }
    from++;

    limit = (unsigned int*)(((char*)dispBitsIndex) + lastWord1InLoop);
    while (from < limit) {
      mask = 0x80000000;
      swapedFrom = *from;
      for (m = 0; m < 32; m++) {
	*to = (mask & swapedFrom) ? 0 : 0xFFFF;
	mask >>= 1;
	to += deltaX;
      }
      toIndex += 32;
      from++;
    }
    mask = 0x80000000;
    swapedFrom = *from;
    for (m = 0; m < 32; m++) {
      if (toIndex >= affectedR) {
        break;
      }
      *to = (mask & swapedFrom) ? 0 : 0xFFFF;
      mask >>= 1;
      to += deltaX;
      toIndex++;
    }
    
    firstWord1 += scanLine1;
    lastWord1InLoop += scanLine1;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits4(int dispBitsIndex, int width, int height, int depth, 
				int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine4 = bytesPerLine(width, 4);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord4 = scanLine4 * affectedT + bytesPerLineRD(affectedL, 4);
  int lastWord4InLoop = scanLine4 * affectedT + bytesPerLineRD(affectedR, 4);
  int line;

  int aL = (affectedL>>3)<<3;
  int deltaX = 1;
  int deltaY = g_gxdp.cbxPitch;

  unsigned char *vramTopPtr;
  int vramIndex;

  int leftStart;
  int maxLine = affectedB;

  if (fullScreenFlag) {
    leftStart = 0;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    leftStart = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top
                : info.rcSipRect.bottom;
  }

  vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  vramIndex = (g_gxdp.cbxPitch) * affectedT + (aL + borderWidthY + leftStart) * (g_gxdp.cBPP/8);

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int*)(((unsigned char*)dispBitsIndex) + firstWord4);
    unsigned int *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;
    unsigned int pixels;

    int i;
    unsigned int mask;
    int shift;

    mask = 0xF0000000;
    shift = 28;
    pixels = *from;
    for (i = 0; i < 8; i++) {
      if (toIndex >= affectedL && toIndex < affectedR) {
	*to = stColors16[(pixels&mask)>>shift];
      }
      to += deltaX;
      toIndex++;
      mask >>= 4;
      shift -= 4;
    }
    from++;

    limit = (unsigned int*)(((unsigned char*)dispBitsIndex) + lastWord4InLoop);

    while (from < limit) {
      pixels = *from;
      mask = 0xF0000000;
      shift = 28;
      for (i = 0; i < 8; i++) {
        *to = stColors16[(pixels&mask)>>shift];
        to += deltaX;
        mask >>= 4;
        shift -= 4;
      }
      toIndex += 8;
      from++;
    }

    pixels = *from;
    mask = 0xF0000000;
    shift = 28;
    for (i = 0; i < 8; i++) {
      if (toIndex >= affectedR) {
	break;
      }
      *to = stColors16[(pixels&mask)>>shift];
      to += deltaX;
      toIndex++;
      mask >>= 4;
      shift -= 4;
    }
    
    firstWord4 += scanLine4;
    lastWord4InLoop += scanLine4;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits8(int dispBitsIndex, int width, int height, int depth, 
				int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine8 = bytesPerLine(width, 8);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord8 = scanLine8 * affectedT + bytesPerLineRD(affectedL, 8);
  int lastWord8InLoop = scanLine8 * affectedT + bytesPerLineRD(affectedR, 8);
  int line;

  int aL = (affectedL>>2)<<2;
  int deltaX = 1;
  int deltaY = g_gxdp.cbxPitch;

  unsigned char *vramTopPtr;
  int vramIndex;

  int leftStart;
  int maxLine = affectedB;

  if (fullScreenFlag) {
    leftStart = 0;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    leftStart = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top
                : info.rcSipRect.bottom;
  }

  vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  vramIndex = (g_gxdp.cbxPitch) * affectedT + (aL + borderWidthY + leftStart) * (g_gxdp.cBPP/8);

  for (line = affectedT; line < maxLine; line++) {
    unsigned char *from = ((unsigned char*)dispBitsIndex) + firstWord8;
    unsigned char *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);

    int toIndex = aL;
    int i;

    for (i = 3; i >= 0; i--) {
      if (toIndex >= affectedL && toIndex < affectedR) {
	*to = stColors16[from[i]];
      }
      to += deltaX;
      toIndex++;
    }
    from += 4;

    limit = ((unsigned char*)dispBitsIndex) + lastWord8InLoop;

    while (from < limit) {
      *to = stColors16[from[3]];
      to += deltaX;
      *to = stColors16[from[2]];
      to += deltaX;
      *to = stColors16[from[1]];
      to += deltaX;
      *to = stColors16[from[0]];
      to += deltaX;
      toIndex += 4;
      from += 4;
    }

    for (i = 3; i >= 0; i--) {
      if (toIndex >= affectedR) {
	break;
      }
      *to = stColors16[from[i]];
      to += deltaX;
      toIndex++;
    }
    
    firstWord8 += scanLine8;
    lastWord8InLoop += scanLine8;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits16(int dispBitsIndex, int width, int height, int depth, 
				 int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord16 = scanLine16 * affectedT + bytesPerLineRD(affectedL, 16);
  int lastWord16InLoop = scanLine16 * affectedT + bytesPerLineRD(affectedR, 16);
  int line;

  int aL = (affectedL>>1)<<1;
  int aR = ((affectedR+1)>>1)<<1;
  int deltaX = 1;
  int deltaY = g_gxdp.cbxPitch;

  unsigned char *vramTopPtr;
  int vramIndex;

  int leftStart;
  int maxLine = affectedB;

  if (fullScreenFlag) {
    leftStart = 0;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    leftStart = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top
                : info.rcSipRect.bottom;
  }

  vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  vramIndex = (g_gxdp.cbxPitch) * affectedT + (aL + borderWidthY + leftStart) * (g_gxdp.cBPP/8);

  for (line = affectedT; line < maxLine; line++) {
    unsigned short *from = (unsigned short *)(((unsigned char*)dispBitsIndex) + firstWord16);
    unsigned short *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;
    unsigned short pixel;

    if (aL != affectedL) {
      to += deltaX;
      pixel = *(from);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      to += deltaX;
      toIndex += 2;
      from += 2;
    }

    limit = (unsigned short *)(((unsigned char*)dispBitsIndex) + lastWord16InLoop);

    while (from < limit) {
      pixel = *(from+1);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      to += deltaX;
      pixel = *(from);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      to += deltaX;
      toIndex += 2;
      from += 2;
    }

    if (aR != affectedR) {
      pixel = *(from+1);
      *to = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
    }

    firstWord16 += scanLine16;
    lastWord16InLoop += scanLine16;
    vramIndex += deltaY;
  }
}

void
iPaqFrameBufferShowDisplayBits32(int dispBitsIndex, int width, int height, int depth, 
				 int affectedL, int affectedR, int affectedT, int affectedB, unsigned char* vram)
{
  int scanLine32 = bytesPerLine(width, 32);
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord32 = scanLine32 * affectedT + bytesPerLineRD(affectedL, 32);
  int lastWord32 = scanLine32 * affectedT + bytesPerLine(affectedR, 32);
  int line;

  int aL = affectedL;
  int deltaX = 1;
  int deltaY = g_gxdp.cbxPitch;

  unsigned char *vramTopPtr;
  int vramIndex;

  int leftStart;
  int maxLine = affectedB;

  if (fullScreenFlag) {
    leftStart = 0;
  } else {
    SIPINFO info;
    info.cbSize = sizeof(info);
    info.dwImDataSize = 0;
    info.pvImData = NULL;
    SHSipInfo(SPI_GETSIPINFO, 0, &info, 0);
    leftStart = info.fdwFlags & SIPF_ON
                ? info.rcSipRect.top
                : info.rcSipRect.bottom;
  }

  vramTopPtr = vram - (g_gxdp.cyHeight - 1) * (g_gxdp.cBPP/8);
  vramIndex = (g_gxdp.cbxPitch) * affectedT + (aL + borderWidthY + leftStart) * (g_gxdp.cBPP/8);

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int *)(((unsigned char*)dispBitsIndex) + firstWord32);
    unsigned int *limit = (unsigned int *)(((unsigned char*)dispBitsIndex) + lastWord32);
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    unsigned int pixel;
    while (from < limit) {
      /*
	0000 0000 1111 1111 0000 0000 0000 0000
	                    1111 1000 0000 0000
	0000 0000 0000 0000 1111 1111 0000 0000
	                    0000 0111 1110 0000
	0000 0000 0000 0000 0000 0000 1111 1111
	                    0000 0000 0001 1111
			    */
      pixel = *from;
      pixel =
	((pixel&0x00F80000)>>8)
	|((pixel&0x0000FC00)>>5)
	|((pixel&0x000000F8)>>3);
      *to = pixel;
      to += deltaX;
      from++;
    }
    firstWord32 += scanLine32;
    lastWord32 += scanLine32;
    vramIndex += deltaY;
  }
}

#  endif


int
ioIpaqFrameBufferShowDisplayBits(int dispBitsIndex, int width, int height, int depth, 
				 int affectedL, int affectedR, int affectedT, int affectedB)
{
  unsigned char *vram;

  /* StartWritingDisplayOnIPAQ(); */
  if (g_gxdp.cBPP != 16) {
    return 0;
  }

  if (!(g_gxdp.ffFormat | kfDirect565)) {
    return 0;
  }

  if ((vram = (unsigned char*)GXBeginDraw()) == NULL) {
    return 0;
  }

  if (depth == 1) {
    iPaqFrameBufferShowDisplayBits1(dispBitsIndex, width, height, depth, 
					  affectedL, affectedR, affectedT, affectedB, vram);
  } else if (depth == 4) {
    iPaqFrameBufferShowDisplayBits4(dispBitsIndex, width, height, depth, 
				    affectedL, affectedR, affectedT, affectedB, vram);
  } else if (depth == 8) {
    iPaqFrameBufferShowDisplayBits8(dispBitsIndex, width, height, depth, 
				    affectedL, affectedR, affectedT, affectedB, vram);
  } else if (depth == 16) {
    iPaqFrameBufferShowDisplayBits16(dispBitsIndex, width, height, depth, 
				     affectedL, affectedR, affectedT, affectedB, vram);
  } else if (depth == 32) {
    iPaqFrameBufferShowDisplayBits32(dispBitsIndex, width, height, depth, 
				     affectedL, affectedR, affectedT, affectedB, vram);
  } else {
    GXEndDraw();
    StopWritingDisplayOnIPAQ();
    return 0;
  }

  GXEndDraw();
  /*StopWritingDisplayOnIPAQ();*/
  return 1;
}



#endif /* USE_GAPI */

int ioShowDisplay(int dispBits, int width, int height, int depth,
		  int affectedL, int affectedR, int affectedT, int affectedB)
{ HDC dc;
  BITMAPINFO *bmi;
  int lines;

  if(!IsWindow(stWindow))
    return 0;

  if(affectedR < affectedL || affectedT > affectedB)
    return 1;

#ifdef USE_GAPI
  if (ioIpaqFrameBufferShowDisplayBits(dispBits, width, height, depth, 
				       affectedL, affectedR, affectedT, affectedB)) {
    return 1;
  }
#endif /* USE_GAPI */

#if 0
  if(ioDirectXShowDisplayBits(dispBits, width, height, depth, 
			      affectedL, affectedR, affectedT, affectedB))
    return 1;
#endif
  /* Try to accumulate the changes */
  if(!updateRightNow)
    {
      updateRect.left = affectedL;
      updateRect.top = affectedT;
      updateRect.right = affectedR;
      updateRect.bottom = affectedB;
      /* Acknowledge the request for deferred updates only 
         if the interpreter is not deferring these by itself */
      if(fDeferredUpdate && !deferDisplayUpdates)
        {
          /* Wait until the next WM_PAINT gets processed */
          InvalidateRect(stWindow,&updateRect,FALSE);
          return 1;
        }
      else
        /* Set the update region manually */
        SetRectRgn(updateRgn, updateRect.left, updateRect.top,
		   updateRect.right,updateRect.bottom);
    }
  else
    {
      /* After a paint, so get affected area from update rect */
      updateRightNow = FALSE;
    }
  if (IsRectEmpty(&updateRect)) return 1; /* Nothing to do */

  affectedL = updateRect.left;
  affectedT = updateRect.top;
  affectedR = updateRect.right;
  affectedB = updateRect.bottom;

  /* Careful here: 
     After resizing the main window the affected area can
     be larger than the area covered by the display bits ... */
  if (affectedR > width) affectedR= width;
  if (affectedB > height) affectedB= height;
  /* ... and don't forget left and top - else reverse_image_* will crash */
  if (affectedL > width) affectedL= width;
  if (affectedT > height) affectedT= height;

  /* Don't draw empty areas */
  if(affectedL == affectedR || affectedT == affectedB) return 1;
  /* reload the update rectangle */
  updateRect.left = affectedL;
  updateRect.top = affectedT;
  updateRect.right = affectedR;
  updateRect.bottom = affectedB;


#if defined(USE_DIB_SECTIONS)
	/******************************************************/
	/* Windows CE version, using DIBSection               */
	/* (does not support palettes or SetDIBitsToDevice()) */
	/******************************************************/
  {
    void* pBits;
    HDC memDC;

    dc = GetDC(stWindow);
    if (!dc)
      error("GetDC() failed");

    memDC = CreateBitmapDC(dc, depth, width, height, &pBits);
    /* Reverse the affected area out of the squeak bitmap, into the DIBSection */

    PROFILE_BEGIN(PROFILE_DISPLAY)
      if( depth < 32 ) {
	if(depth == 16)
	  reverse_image_words((unsigned int*) pBits, (unsigned int*) dispBits,
			      depth, width, &updateRect);
	else
	  reverse_image_bytes((unsigned int*) pBits, (unsigned int*) dispBits,
			      depth, width, &updateRect);
      } else {
	copy_image_words((unsigned int*)pBits, (unsigned int*) dispBits, 
			 depth, width, &updateRect);
      }
    PROFILE_END(ticksForReversal)
      
      PROFILE_BEGIN(PROFILE_DISPLAY);
    BitBlt(dc, 
	   updateRect.left,/* dst_x */
	   updateRect.top, /* dst_y */
	   (updateRect.right - updateRect.left),/* dst_w */
	   (updateRect.bottom - updateRect.top),/* dst_h */
	   memDC,
	   updateRect.left, /* src_x */
	   updateRect.top,  /* src_y */
	   SRCCOPY);

    ReleaseBitmapDC(memDC);
    ReleaseDC(stWindow,dc);
    PROFILE_END(ticksForBlitting);
  }
#else /* !defined(USE_DIB_SECTIONS) */

  bmi = BmiForDepth(depth);
  if(!bmi)
    {
      abortMessage(TEXT("Aborting Squeak!!!!\nColor depth %d not supported"),depth);
    }

  /* reverse the image bits if necessary */
#ifndef NO_BYTE_REVERSAL
  PROFILE_BEGIN(PROFILE_DISPLAY)
  if( depth < 32 )
    if(depth == 16)
      reverse_image_words((unsigned int*) dispBits, (unsigned int*) dispBits,
			  depth, width, &updateRect);
    else
      reverse_image_bytes((unsigned int*) dispBits, (unsigned int*) dispBits,
			  depth, width, &updateRect);
  PROFILE_END(ticksForReversal)
#endif NO_BYTE_REVERSAL

  bmi->bmiHeader.biWidth = width;
  bmi->bmiHeader.biHeight = -height;
  bmi->bmiHeader.biSizeImage = 0;

  dc = GetDC(stWindow);
  if(!dc) {
    printLastError(TEXT("ioShowDisplayBits: GetDC() failed"));
    return 0;
  }

  SelectPalette(dc,palette,0);
  RealizePalette(dc);


  PROFILE_BEGIN(PROFILE_DISPLAY)
  /* use the actual affected region rather than 
     the complete recangle for clipping */
  SelectClipRgn(dc,updateRgn);

  lines = SetDIBitsToDevice(dc,
  	    0, /* dst_x */
  	    0, /* dst_y */
  	    width,  /* dst_w */
  	    height, /* dst_h */
  	    0, /* src_x */
  	    0, /* src_y */
  	    0, /* start scan line in DIB */
  	    height, /* num scan lines in DIB */
	    (void*) dispBits,  /* bits */
	    bmi,
	    DIB_RGB_COLORS);

  if(lines == 0) {
    /* Note: the above is at least five times faster than what follows.
       Unfortunately, it also requires quite a bit of resources to
       be available. These are almost always available except in a
       few extreme conditions - but to compensate for those the
       following is provided. */
    int pitch, start, end, nPix, line, left;
    int bitsPtr;

    /* compute pitch of form */
    pitch = ((width * depth) + 31 & ~31) / 8;
    /* compute first word of update region */
    start = ((updateRect.left * depth) & ~31) / 8;
    /* compute last word of update region */
    end   = ((updateRect.right * depth) + 31 & ~31) / 8;
    /* compute #of bits covered in update region */
    nPix = ((end - start) * 8) / depth;
    left = (start * 8) / depth;
    bmi->bmiHeader.biWidth = nPix;
    bmi->bmiHeader.biHeight = 1;
    bmi->bmiHeader.biSizeImage = 0;
    bitsPtr = dispBits + start + (updateRect.top * pitch);
    for(line = updateRect.top; line < updateRect.bottom; line++) {
      lines = SetDIBitsToDevice(dc, left, line, nPix, 1, 0, 0, 0, 1,
				(void*) bitsPtr, bmi, DIB_RGB_COLORS);
      bitsPtr += pitch;
    }
  }

  SelectClipRgn(dc,NULL);
  PROFILE_END(ticksForBlitting)

  ReleaseDC(stWindow,dc);

  if(lines == 0) {
    printLastError(TEXT("SetDIBitsToDevice failed"));
    warnPrintf(TEXT("width=%d,height=%d,bits=%X,dc=%X\n"), 
	       width, height, dispBits,dc);
  }
  /* reverse the image bits if necessary */
#ifndef NO_BYTE_REVERSAL
  PROFILE_BEGIN(PROFILE_DISPLAY)
  if( depth < 32 )
    if(depth == 16)
      reverse_image_words((unsigned int*) dispBits, (unsigned int*) dispBits,
			  depth, width, &updateRect);
    else
      reverse_image_bytes((unsigned int*) dispBits, (unsigned int*) dispBits,
			  depth, width, &updateRect);
  PROFILE_END(ticksForReversal)
#endif /* NO_BYTE_REVERSAL */
#endif /* defined(_WIN32_WCE) */
  return 1;
}

/****************************************************************************/
/*                      Clipboard                                           */
/****************************************************************************/

int clipboardSize(void)
{ HANDLE h;
  unsigned char *src;
  int len,tel;

  if(!IsClipboardFormatAvailable(CF_TEXT)) /* check for format CF_TEXT */
    return 0;
  if(!OpenClipboard(stWindow))
    return 0;
  h = GetClipboardData(CF_TEXT);
  src = (unsigned char*)GlobalLock(h);
  len = src ? strlen((const char*)src) : 0;
  /* RvL 17-04-1998
     Look for CrLf sequences */
  tel = len;
  while(tel--)
    if((*src++ == 0x0d) && (*src == 0x0a))
      len--;
  GlobalUnlock(h);
  CloseClipboard();
  return len;
}
    
/* send the given string to the clipboard */
int clipboardWriteFromAt(int count, int byteArrayIndex, int startIndex)
{ HANDLE h;
  unsigned char *dst,*src;
  int adjCount, tel;

  if(!OpenClipboard(stWindow))
    return 0;

  /* RvL 17-04-1998
     adjust count for inserting 0A into clipboard string */
  src = (unsigned char *)byteArrayIndex + startIndex;
  tel = count;
  adjCount = 0;
  while(tel--)
    if ((*src++ == 0x0D) && (*src != 0x0A))
      adjCount++;

  h = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, count+1+adjCount);
  dst = (unsigned char*)GlobalLock(h);
  src = (unsigned char *)byteArrayIndex + startIndex;
  /* Remap from Mac Roman to Ansi Latin */
  while(count--)
    if((*src == 0x0D) && (src[1] != 0x0A))
      { /* special case: crlf translation */
        *dst++ = *src++;
        *dst++ = 0x0A;
      }
    else
      { /* regular case: lookup translation */
        *dst++ = iKeymap[*src++];
      }
  *dst = 0;
  EmptyClipboard();
  SetClipboardData(CF_TEXT, h);
  /* Note: After setting clipboard data, 
     the memory block previously allocated belongs to
     the clipboard - not to the app. */
  CloseClipboard();
  return 1;
}


/* transfer the clipboard data into the given byte array */
int clipboardReadIntoAt(int count, int byteArrayIndex, int startIndex)
{ HANDLE h;
  unsigned char *src,*dst;
  int len;

  if(!IsClipboardFormatAvailable(CF_TEXT)) /* check for format CF_TEXT */
    return 0;
  if(!OpenClipboard(stWindow))
    return 0;
  dst= (unsigned char *)byteArrayIndex + startIndex;
  h = GetClipboardData(CF_TEXT);
  src = (unsigned char*)GlobalLock(h);
  len = src ? strlen((const char*)src) : 0;
  if(len > count) len = count;
  /* Remap from Ansi Latin to Mac Roman */
  while(len--)
    if((*src == 0x0D) && (src[1] == 0x0A))
      { /* RvL 17-04-1998
           drop the line feed after a carriage return
           but leave lone line feeds alone
           may transfer less than 'len' bytes but who cares
           i.e. if they use the clipboardSize there should
           be no problem.*/
        *dst++ = *src++;
        src++;
      }
    else
      { /* regular case: lookup translation */
        *dst++ = keymap[*src++];
      }
  GlobalUnlock(h);
  CloseClipboard();
  return len;
}


/****************************************************************************/
/*                          Profiling                                       */
/****************************************************************************/
#ifndef PROFILE
int clearProfile(void) { return 1;}
int dumpProfile(void) {return 1;}
int startProfiling(void) {return 1;}
int stopProfiling(void) {return 1;}
#endif

/****************************************************************************/
/*                    Image / VM File Naming                                */
/****************************************************************************/

int vmPathSize(void)
{
  return lstrlen(vmPath);
}

int vmPathGetLength(int sqVMPathIndex, int length)
{
  char *stVMPath= (char *)sqVMPathIndex;
  int count, i;

  count= lstrlen(vmPath);
  count= (length < count) ? length : count;

  /* copy the file name into the Squeak string */
#ifdef UNICODE
# if _WIN32_WCE >= 300
  {
    char *s;
    s = fromUnicode(vmPath);
    for (i = 0; i < count; i++)
      stVMPath[i] = s[i];
    // in Squeak, aString is not null-terminated.
    // strcpy(stVMPath,fromUnicode(vmPath));
 }
# else
  for (i= 0; i < count; i++)
    stVMPath[i]= (char) vmPath[i]; /* will remove leading zeros from unicode */
# endif
#endif
  return count;
}

int imageNameSize(void)
{
  return strlen(imageName);
}

int imageNameGetLength(int sqImageNameIndex, int length)
{
  char *sqImageName= (char *)sqImageNameIndex;
  int count, i;

  count= strlen(imageName);
  count= (length < count) ? length : count;

  /* copy the file name into the Squeak string */
  for (i= 0; i < count; i++)
    sqImageName[i]= imageName[i];

  return count;
}

int imageNamePutLength(int sqImageNameIndex, int length)
{
  char *sqImageName= (char *)sqImageNameIndex;
  char tmpImageName[MAX_PATH+1];
  char *tmp;
  int count, i;

  count= (IMAGE_NAME_SIZE < length) ? IMAGE_NAME_SIZE : length;

  /* copy the file name into a null-terminated C string */
  for (i= 0; i < count; i++)
    tmpImageName[i]= sqImageName[i];

  tmpImageName[count]= 0;

  /* Note: We have to preserve the fully qualified image path */
  tmp = strrchr(tmpImageName,'\\');
  if(tmp)
    { /* fully qualified */
      strcpy(imageName,tmpImageName);
    }
  else
    { /* not qualified */
      tmp = strrchr(imageName,'\\');
      if(!tmp)
        strcpy(imageName,tmpImageName);
      else
        {
          tmp++; *tmp = 0;
          count = IMAGE_NAME_SIZE - (tmp-imageName);
          if(count < length)
            tmpImageName[count] = 0;
          strcat(imageName,tmpImageName);
        }
    }
  SetWindowTitle();
  return 1;
}

/****************************************************************************/
/*                      System Attributes                                   */
/****************************************************************************/
char * GetAttributeString(int id) {
	/* This is a hook for getting various status strings back from
	   the OS. In particular, it allows Squeak to be passed arguments
	   such as the name of a file to be processed. Command line options
	   could be reported this way as well.
	*/
  /* id == 0 : return the full name of the VM */
  if(id == 0) return fromUnicode(vmName);
  /* 0 < id <= 1000 : return options of the image (e.g. given *after* the image name) */
  if(id > 0 && id <= 1000)
    return GetImageOption(id-1);
  /* id < 0 : return options of the VM (e.g. given *before* the image name) */
  if(id < 0)
    return GetVMOption(0-id-1);
  /* special attributes */
  if(id == 1001) /* Primary OS type */
      return WIN32_NAME;
  if(id == 1002) /* Secondary OS type */
      return WIN32_OS_NAME;
  if(id == 1003) /* Processor type */
      return WIN32_PROCESSOR_NAME;
  if(id == 1004) /* Interpreter version */
	  return (char*) interpreterVersion;
  return "";
}

int attributeSize(int id) {
	return strlen(GetAttributeString(id));
}

int getAttributeIntoLength(int id, int byteArrayIndex, int length) {
  char *srcPtr, *dstPtr, *end;
  int charsToMove;
  
  srcPtr = GetAttributeString(id);
  charsToMove = strlen(srcPtr);
  if (charsToMove > length) {
    charsToMove = length;
  }

  dstPtr = (char *) byteArrayIndex;
  end = srcPtr + charsToMove;
  while (srcPtr < end) {
    *dstPtr++ = *srcPtr++;
  }
  return charsToMove;
}


/****************************************************************************/
/*                      File Startup                                        */
/****************************************************************************/

static TCHAR *lstrrchr(TCHAR *source, TCHAR c)
{ TCHAR *tmp;

  /* point to the last char */
  tmp = source + lstrlen(source)-1;
  while(tmp >= source)
    if(*tmp == c) return tmp;
    else tmp--;
  return NULL;
}

/* Check if the path/file name is subdirectory of the image path */
int isLocalFileName(TCHAR *fileName)
{
  int i;
  for(i=0; i<lstrlen(imagePath); i++)
    if(imagePath[i] != fileName[i]) return 0;
  return 1;
}

#if defined(_WIN32_WCE)
	/* WinCE does not support short file names, and has
	   no concept of a current directory. Space is at a
	   premium, the file system is small, and we are unlikely
	   to have a full sources file anyway (too big). All these
	   factors means that we stick with a simpler scheme, of
	   either requiring the image name to be fully pathed, or
	   if not, popping up a file open dialog */

void SetupFilesAndPath()
{ TCHAR *tmp;
  
  /* get the VM directory */
  tmp = lstrrchr(vmPath,U_BACKSLASH[0]);
  if(tmp) tmp[1] = 0;

  lstrcpy(imagePath, toUnicode(imageName));
  tmp = lstrrchr(imagePath,U_BACKSLASH[0]);
  if(tmp) tmp[1] = 0;
}

#else /* defined(_WIN32_WCE) */

void
LongFileNameFromPossiblyShortName(TCHAR *nameBuffer)
{ TCHAR oldDir[MAX_PATH+1];
  TCHAR testName[13];
  TCHAR nameBuf[MAX_PATH+1];
  TCHAR *shortName;
  WIN32_FIND_DATA findData;
  HANDLE findHandle;

  GetCurrentDirectory(MAX_PATH,oldDir);
  shortName = lstrrchr(nameBuffer,U_BACKSLASH[0]);
  if(!shortName) shortName = lstrrchr(nameBuffer,U_SLASH[0]);
  if(!shortName) return;
  /* if the file name is longer than 8.3
     this can't be a short name */
  *(shortName++) = 0;
  if(lstrlen(shortName) > 12)
    goto notFound;

  /* back up the old and change to the given directory,
     this makes searching easier */
  lstrcpy(nameBuf, nameBuffer);
  lstrcat(nameBuf,TEXT("\\"));
  SetCurrentDirectory(nameBuf);

  /* now search the directory */
  findHandle = FindFirstFile(TEXT("*.*"),&findData);
  if(findHandle == INVALID_HANDLE_VALUE) goto notFound; /* nothing found */
  do {
    if(lstrcmp(findData.cFileName,TEXT("..")) && lstrcmp(findData.cFileName,TEXT(".")))
      lstrcpy(testName,findData.cAlternateFileName);
    else
      *testName = 0;
    if(lstrcmp(testName,shortName) == 0) /* gotcha! */
      {
        FindClose(findHandle);
        /* recurse down */
        lstrcpy(nameBuf, findData.cFileName);
        goto recurseDown;
      }
  } while(FindNextFile(findHandle,&findData) != 0);
  /* nothing appropriate found */
  FindClose(findHandle);
notFound:
  lstrcpy(nameBuf, shortName);
recurseDown:
  /* recurse down */
  LongFileNameFromPossiblyShortName(nameBuffer);
  lstrcat(nameBuffer,TEXT("\\"));
  lstrcat(nameBuffer,nameBuf);
  SetCurrentDirectory(oldDir);
}

void
SetupFilesAndPath()
{ TCHAR *tmp;
  TCHAR tmpName[MAX_PATH+1];
  TCHAR imageNameW[MAX_PATH+1];

  /* get the full path for the image */
  lstrcpy(tmpName, toUnicode(imageName));
  GetFullPathName(tmpName,MAX_PATH,imageNameW,&tmp);
  /* check if it is a short name and if so convert 
     it to the full name */
  LongFileNameFromPossiblyShortName(imageNameW);

  /* and copy back to a C string */
  strcpy(imageName, fromUnicode(imageNameW));

  /* get the VM directory */
  lstrcpy(vmPath, vmName);
  tmp = lstrrchr(vmPath,U_BACKSLASH[0]);
  if(tmp) *tmp = 0;
  else GetCurrentDirectory(MAX_PATH,vmPath);
  lstrcat(vmPath,U_BACKSLASH);

  lstrcpy(imagePath, imageName);
  tmp = lstrrchr(imagePath,U_BACKSLASH[0]);
  if(tmp) tmp[1] = 0;
}

#endif /* !defined(_WIN32_WCE) */

/* SqueakImageLength(): 
   Return the length of the image if it is a valid Squeak image file.
   Otherwise return 0. */
DWORD SqueakImageLength(TCHAR *fileName)
{ DWORD dwRead, dwSize, magic = 0;
  HANDLE hFile;

  /* open image file */
  hFile = CreateFile(fileName, GENERIC_READ, FILE_SHARE_READ, 
                     NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if(hFile == INVALID_HANDLE_VALUE) 
    return 0;
  /* read magic number */
  if(!ReadFile(hFile, &magic, 4, &dwRead, NULL))
    {
      CloseHandle(hFile);
      return 0;
    }
  /* get the image's size */
  dwSize = GetFileSize(hFile, NULL);
  CloseHandle(hFile);
  if(!readableFormat(magic) && !readableFormat(byteSwapped(magic)))
    return 0;
  else
    return dwSize;
}

/****************************************************************************/
/*                      Startup helper functions                            */
/****************************************************************************/

/* findImageFile():
   Search the current directory for exactly one image file.
   If it is found, copy the name into imageName and return true.
*/
int findImageFile(void)
{
  WIN32_FIND_DATA findData;
  HANDLE findHandle;
  int nextFound;

  findHandle = FindFirstFile(TEXT("*.image"),&findData);
  if(findHandle == INVALID_HANDLE_VALUE)
	  return 0; /* Not found */
  nextFound = FindNextFile(findHandle,&findData);
  FindClose(findHandle);
  if(nextFound) return 0; /* more than one entry */
  strcpy(imageName, fromUnicode(findData.cFileName));
  return 1;
}

/* openImageFile():
   Pop up a file open dialog for image files.
   Copy the selection into imageName and return true.
*/
int openImageFile(void)
{
  OPENFILENAME ofn;
  memset(&ofn, 0, sizeof(ofn));
  ofn.lStructSize = sizeof(ofn);
  ofn.lpstrFilter = TEXT("Image Files (*.image)\0*.image\0All Files (*.*)\0*.*\0");
  ofn.lpstrFile = vmPath;
  ofn.nMaxFile = MAX_PATH;
  ofn.lpstrTitle = TEXT("SQUEAK: Please select an image file...");
  ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
  ofn.lpstrDefExt = TEXT("image");
  MessageBeep(MB_ICONSTOP);
  if (!GetOpenFileName(&ofn)) return 0;
  strcpy(imageName, fromUnicode(vmPath));
  return 1;
}

/****************************************************************************/
/*                      Usage of Squeak                                     */
/****************************************************************************/

/* print usage with different output levels */
int printUsage(int level)
{
  switch(level) {
    case 0: /* No command line given */
      abortMessage(TEXT("Usage: squeak [options] <imageFile>"));
      break;
    case 1: /* full usage */
      abortMessage(TEXT("%s"),
                   TEXT("Usage: squeak [vmOptions] imageFile [imageOptions]\n\n")
                   TEXT("vmOptions:\n")
                   TEXT("\t-service: ServiceName \t(install Squeak as NT service)\n")
                   TEXT("\t-headless \t\t(force Squeak to run headless)\n")
                   TEXT("\t-log: LogFile \t\t(use LogFile for VM messages)\n")
                   TEXT("\t-memory: megaByte \t(set memory to megaByte MB)"));
      break;
    case 2: /* No image found */
    default:
      abortMessage(
        TEXT("Could not open the Squeak image file '%s'\n\n")
        TEXT("There are several ways to open a Squeak image file. You can:\n")
        TEXT("  1. Double-click on the desired image file.\n")
        TEXT("  2. Drop the image file onto the Squeak application.\n")
        TEXT("Aborting...\n"), toUnicode(imageName));
  }
  return -1;
}

#ifdef USE_GAPI

void
OpenInputOnIPAQ()
{
# ifndef LANDSCAPE_SCREEN
  g_gxkl = GXGetDefaultKeys(GX_NORMALKEYS);
# else
  g_gxkl = GXGetDefaultKeys(GX_LANDSCAPEKEYS);
# endif
  GXOpenInput();
}

void
OpenDisplayOnIPAQ()
{
  if (GXOpenDisplay(stWindow, GX_FULLSCREEN) == 0) {
    g_gxdp.cBPP = 0;
  } else {
    g_gxdp = GXGetDisplayProperties();
    OpenInputOnIPAQ();
  }
}

void
CloseDisplayOnIPAQ()
{
  GXCloseDisplay();
}

void
StartWritingDisplayOnIPAQ()
{
  if (GXOpenDisplay(stWindow, GX_FULLSCREEN) == 0) {
    g_gxdp.cBPP = 0;
  } else {
    g_gxdp = GXGetDisplayProperties();
  }
}

void
StopWritingDisplayOnIPAQ()
{
  GXCloseDisplay();
}

#endif
